# distutils: libraries = NTL_LIBRARIES gmp m
# distutils: extra_compile_args = NTL_CFLAGS
# distutils: include_dirs = NTL_INCDIR
# distutils: library_dirs = NTL_LIBDIR
# distutils: extra_link_args = NTL_LIBEXTRA
# distutils: language = c++

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from cysignals.signals cimport sig_on, sig_off
from sage.ext.cplusplus cimport ccrepr, ccreadstr

include 'misc.pxi'
include 'decl.pxi'

from cpython.object cimport Py_EQ, Py_NE
from sage.rings.integer import Integer
from sage.rings.integer_ring import IntegerRing
from sage.rings.integer cimport Integer
from sage.libs.ntl.ntl_ZZ cimport ntl_ZZ
from sage.rings.rational cimport Rational

from sage.libs.ntl.ntl_ZZ import unpickle_class_args
from sage.libs.ntl.convert cimport PyLong_to_ZZ, mpz_to_ZZ

from sage.libs.ntl.ntl_ZZ_pContext cimport ntl_ZZ_pContext_class
from sage.libs.ntl.ntl_ZZ_pContext import ntl_ZZ_pContext

from sage.misc.randstate cimport current_randstate


ZZ_sage = IntegerRing()


def ntl_ZZ_p_random_element(v):
    """
    Return a random number modulo p.

    EXAMPLES::

        sage: a = sage.libs.ntl.ntl_ZZ_p.ntl_ZZ_p_random_element(17)
        sage: type(a)
        <class 'sage.libs.ntl.ntl_ZZ_p.ntl_ZZ_p'>
        sage: a.modulus()
        17
    """
    current_randstate().set_seed_ntl(False)

    cdef ntl_ZZ_p y
    v = ntl_ZZ_pContext(v)
    y = ntl_ZZ_p(0,v)
    sig_on()
    ZZ_p_random(y.x)
    sig_off()
    return y


##############################################################################
#
# ZZ_p_c: integers modulo p
#
##############################################################################
cdef class ntl_ZZ_p():
    r"""
    The \class{ZZ_p} class is used to represent integers modulo `p`.
    The modulus `p` may be any positive integer, not necessarily prime.

    Objects of the class \class{ZZ_p} are represented as a \code{ZZ} in the
    range `0, \ldots, p-1`.

    Each \class{ZZ_p} contains a pointer of a \class{ZZ_pContext} which
    contains pre-computed data for NTL.  These can be explicitly constructed
    and passed to the constructor of a \class{ZZ_p} or the \class{ZZ_pContext}
    method \code{ZZ_p} can be used to construct a \class{ZZ_p} element.

    This class takes care of making sure that the C++ library NTL global
    variable is set correctly before performing any arithmetic.
    """

    def __init__(self, v=None, modulus=None):
        r"""
        Initialize an NTL integer mod p.

        EXAMPLES::

            sage: c = ntl.ZZ_pContext(11)
            sage: ntl.ZZ_p(12r, c)
            1
            sage: ntl.ZZ_p(Integer(95413094), c)
            7
            sage: ntl.ZZ_p('-1', c)
            10

        AUTHOR: Joel B. Mohler (2007-06-14)
        """
        if modulus is None:
            raise ValueError("You must specify a modulus when creating a ZZ_p.")

        # self.c._assert_is_current_modulus()  # The context was restored in __new__

        cdef ZZ_c temp, num, den
        if v is not None:
            if isinstance(v, ntl_ZZ_p):
                self.x = (<ntl_ZZ_p>v).x
            elif isinstance(v, int):
                PyLong_to_ZZ(&temp, v)
                self.x = ZZ_to_ZZ_p(temp)
            elif isinstance(v, Integer):
                mpz_to_ZZ(&temp, (<Integer>v).value)
                self.x = ZZ_to_ZZ_p(temp)
            elif isinstance(v, Rational):
                mpz_to_ZZ(&num, (<Integer>v.numerator()).value)
                mpz_to_ZZ(&den, (<Integer>v.denominator()).value)
                ZZ_p_div(self.x, ZZ_to_ZZ_p(num), ZZ_to_ZZ_p(den))
            else:
                str_v = str(v)  # can cause modulus to change; Issue #25790
                self.c.restore_c()
                ccreadstr(self.x, str_v)

    def __cinit__(self, v=None, modulus=None):
        #################### WARNING ###################
        ## Before creating a ZZ_p, you must create a  ##
        ## ZZ_pContext, and restore it.  In Python,   ##
        ## the error checking in __init__ will prevent##
        ## you from constructing an ntl_ZZ_p          ##
        ## inappropriately.  However, from Cython, you##
        ## could do r = ntl_ZZ_p.__new__(ntl_ZZ_p) without
        ## first restoring a ZZ_pContext, which could ##
        ## have unfortunate consequences.  See _new  ##
        ## defined below for an example of the right  ##
        ## way to short-circuit __init__ (or just call##
        ## _new in your own code).                    ##
        ################################################
        if modulus is None:
            return
        if isinstance(modulus, ntl_ZZ_pContext_class):
            self.c = <ntl_ZZ_pContext_class>modulus
            self.c.restore_c()
        else:
            self.c = <ntl_ZZ_pContext_class>ntl_ZZ_pContext(modulus)
            self.c.restore_c()

    cdef ntl_ZZ_p _new(self):
        cdef ntl_ZZ_p r
        self.c.restore_c()
        r = ntl_ZZ_p.__new__(ntl_ZZ_p)
        r.c = self.c
        return r

    def __reduce__(self):
        """
        EXAMPLES::

            sage: a = ntl.ZZ_p(4,7)
            sage: loads(dumps(a)) == a
            True
        """
        return unpickle_class_args, (ntl_ZZ_p, (self.lift(), self.modulus_context()))

    def modulus_context(self):
        """
        Return the modulus for ``self``.

        EXAMPLES::

            sage: x = ntl.ZZ_p(5,17)
            sage: c = x.modulus_context()
            sage: y = ntl.ZZ_p(3,c)
            sage: x+y
            8
            sage: c == y.modulus_context()
            True
            sage: c == ntl.ZZ_p(7,17).modulus_context()
            True
        """
        return self.c

    def __repr__(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: ntl.ZZ_p(7,192).__repr__()
            '7'
        """
        self.c.restore_c()
        return ccrepr(self.x)

    def __richcmp__(ntl_ZZ_p self, other, int op):
        r"""
        Compare ``self`` to ``other``.

        EXAMPLES::

            sage: c = ntl.ZZ_pContext(11)
            sage: ntl.ZZ_p(12r, c) == ntl.ZZ_p(1, c)
            True
            sage: ntl.ZZ_p(35r, c) == ntl.ZZ_p(1, c)
            False
            sage: "2" == ntl.ZZ_p(35r, c)
            True
            sage: ntl.ZZ_p(35r, c) == 2
            True
        """
        self.c.restore_c()

        if op != Py_EQ and op != Py_NE:
            raise TypeError("integers mod p are not ordered")

        cdef ntl_ZZ_p b
        try:
            b = <ntl_ZZ_p?>other
        except TypeError:
            b = ntl_ZZ_p(other, self.c)

        return (op == Py_EQ) == (self.x == b.x)

    def __invert__(ntl_ZZ_p self):
        r"""
        EXAMPLES::

            sage: c=ntl.ZZ_pContext(11)
            sage: ~ntl.ZZ_p(2r,modulus=c)
            6
        """
        cdef ntl_ZZ_p r = self._new()
        sig_on()
        self.c.restore_c()
        ZZ_p_inv(r.x, self.x)
        sig_off()
        return r

    def __mul__(ntl_ZZ_p self, other):
        """
        EXAMPLES::

            sage: x = ntl.ZZ_p(5,31) ; y = ntl.ZZ_p(8,31)
            sage: x*y ## indirect doctest
            9
        """
        cdef ntl_ZZ_p y
        cdef ntl_ZZ_p r = self._new()
        if not isinstance(other, ntl_ZZ_p):
            other = ntl_ZZ_p(other,self.c)
        elif self.c is not (<ntl_ZZ_p>other).c:
            raise ValueError("You cannot perform arithmetic with elements of different moduli.")
        y = other
        self.c.restore_c()
        ZZ_p_mul(r.x, self.x, y.x)
        return r

    def __sub__(ntl_ZZ_p self, other):
        """
        EXAMPLES::

            sage: x = ntl.ZZ_p(5,31) ; y = ntl.ZZ_p(8,31)
            sage: x-y ## indirect doctest
            28
            sage: y-x
            3
        """
        if not isinstance(other, ntl_ZZ_p):
            other = ntl_ZZ_p(other,self.c)
        elif self.c is not (<ntl_ZZ_p>other).c:
            raise ValueError("You cannot perform arithmetic with elements of different moduli.")
        cdef ntl_ZZ_p r = self._new()
        self.c.restore_c()
        ZZ_p_sub(r.x, self.x, (<ntl_ZZ_p>other).x)
        return r

    def __add__(ntl_ZZ_p self, other):
        """
        EXAMPLES::

            sage: x = ntl.ZZ_p(5,31) ; y = ntl.ZZ_p(8,31)
            sage: x+y ## indirect doctest
            13
        """
        cdef ntl_ZZ_p y
        cdef ntl_ZZ_p r = ntl_ZZ_p(modulus=self.c)
        if not isinstance(other, ntl_ZZ_p):
            other = ntl_ZZ_p(other,modulus=self.c)
        elif self.c is not (<ntl_ZZ_p>other).c:
            raise ValueError("You cannot perform arithmetic with elements of different moduli.")
        y = other
        sig_on()
        self.c.restore_c()
        ZZ_p_add(r.x, self.x, y.x)
        sig_off()
        return r

    def __neg__(ntl_ZZ_p self):
        """
        EXAMPLES::

            sage: x = ntl.ZZ_p(5,31)
            sage: -x ## indirect doctest
            26
        """
        cdef ntl_ZZ_p r = ntl_ZZ_p(modulus=self.c)
        sig_on()
        self.c.restore_c()
        ZZ_p_negate(r.x, self.x)
        sig_off()
        return r

    def __pow__(ntl_ZZ_p self, long e, ignored):
        """
        EXAMPLES::

            sage: x = ntl.ZZ_p(5,31)
            sage: x**3 ## indirect doctest
            1
        """
        cdef ntl_ZZ_p r = ntl_ZZ_p(modulus=self.c)
        sig_on()
        self.c.restore_c()
        ZZ_p_power(r.x, self.x, e)
        sig_off()
        return r

    def __int__(self):
        """
        Return ``self`` as an int.

        EXAMPLES::

            sage: x = ntl.ZZ_p(3,8)
            sage: x.__int__()
            3
            sage: type(x.__int__())
            <... 'int'>
        """
        return self.get_as_int()

    cdef int get_as_int(ntl_ZZ_p self) noexcept:
        r"""
        Return value as C int.
        Return value is only valid if the result fits into an int.

        AUTHOR: David Harvey (2006-08-05)
        """
        self.c.restore_c()
        return ZZ_p_to_int(self.x)

    def _get_as_int_doctest(self):
        r"""
        This method exists solely for automated testing of get_as_int().

        EXAMPLES::

            sage: c = ntl.ZZ_pContext(20)
            sage: x = ntl.ZZ_p(42,modulus=c)
            sage: i = x._get_as_int_doctest()
            sage: i
            2
            sage: type(i)
            <... 'int'>
        """
        self.c.restore_c()
        return self.get_as_int()

    cdef void set_from_int(ntl_ZZ_p self, int value) noexcept:
        r"""
        Set the value from a C int.

        AUTHOR: David Harvey (2006-08-05)
        """
        self.c.restore_c()
        self.x = int_to_ZZ_p(value)

    def _set_from_int_doctest(self, value):
        r"""
        This method exists solely for automated testing of set_from_int().

        EXAMPLES::

            sage: c=ntl.ZZ_pContext(ntl.ZZ(20))
            sage: x = ntl.ZZ_p(modulus=c)
            sage: x._set_from_int_doctest(42)
            sage: x
            2
            sage: x = ntl.ZZ_p(7,81)
            sage: x._set_from_int_doctest(int(3))
            sage: x
            3
        """
        self.c.restore_c()
        self.set_from_int(int(value))

    def lift(self):
        """
        Return a lift of ``self`` as an ntl.ZZ object.

        EXAMPLES::

            sage: x = ntl.ZZ_p(8,18)
            sage: x.lift()
            8
            sage: type(x.lift())
            <class 'sage.libs.ntl.ntl_ZZ.ntl_ZZ'>
        """
        cdef ntl_ZZ r = ntl_ZZ()
        self.c.restore_c()
        r.x = ZZ_p_rep(self.x)
        return r

    def modulus(self):
        r"""
        Return the modulus as an NTL ZZ.

        EXAMPLES::

            sage: c = ntl.ZZ_pContext(ntl.ZZ(20))
            sage: n = ntl.ZZ_p(2983,c)
            sage: n.modulus()
            20
        """
        cdef ntl_ZZ r = ntl_ZZ()
        self.c.restore_c()
        ZZ_p_modulus( &r.x, &self.x )
        return r

    def lift_centered(self):
        """
        Compute a representative of ``self`` in `(-n/2 , n/2]` as an
        ``ntl.ZZ`` object.

        OUTPUT:

        - A ``ntl.ZZ`` object `r` such that  `-n/2 < r \\leq n/2` and `Mod(r, n) == self`.

        EXAMPLES::

            sage: x = ntl.ZZ_p(8, 18)
            sage: x.lift_centered()
            8
            sage: type(x.lift_centered())
            <class 'sage.libs.ntl.ntl_ZZ.ntl_ZZ'>
            sage: x = ntl.ZZ_p(12, 18)
            sage: x.lift_centered()
            -6
            sage: type(x.lift_centered())
            <class 'sage.libs.ntl.ntl_ZZ.ntl_ZZ'>
        """
        cdef ntl_ZZ r = self.lift()
        cdef ntl_ZZ m = self.modulus()
        if r*2 > m:
            r -= m
        return r

    def _integer_(self, ZZ=None):
        """
        Return a lift of ``self`` as a Sage integer.

        EXAMPLES::

            sage: x = ntl.ZZ_p(8,188)
            sage: x._integer_()
            8

            sage: type(x._integer_())
            <class 'sage.rings.integer.Integer'>
        """
        self.c.restore_c()
        cdef ZZ_c rep = ZZ_p_rep(self.x)
        cdef Integer ans = Integer.__new__(Integer)
        ZZ_to_mpz(ans.value, &rep)
        return ans

    def _sage_(self):
        r"""
        Return the value as a sage IntegerModRing.

        EXAMPLES::

            sage: c = ntl.ZZ_pContext(20)
            sage: n = ntl.ZZ_p(2983, c)
            sage: type(n._sage_())
            <class 'sage.rings.finite_rings.integer_mod.IntegerMod_int'>
            sage: n
            3

        AUTHOR: Joel B. Mohler
        """
        from sage.rings.finite_rings.integer_mod_ring import IntegerModRing

        cdef ZZ_c rep
        self.c.restore_c()
        rep = ZZ_p_rep(self.x)
        cdef Integer ans = Integer.__new__(Integer)
        ZZ_to_mpz(ans.value, &rep)
        return IntegerModRing(self.modulus()._integer_())(ans)



# sage.doctest: needs sage.libs.pari
"""
The cuspidal subspace

EXAMPLES::

    sage: S = CuspForms(SL2Z,12); S
    Cuspidal subspace of dimension 1 of Modular Forms space of dimension 2 for
    Modular Group SL(2,Z) of weight 12 over Rational Field
    sage: S.basis()
    [q - 24*q^2 + 252*q^3 - 1472*q^4 + 4830*q^5 + O(q^6)]

    sage: S = CuspForms(Gamma0(33),2); S
    Cuspidal subspace of dimension 3 of Modular Forms space of dimension 6 for
    Congruence Subgroup Gamma0(33) of weight 2 over Rational Field
    sage: S.basis()
    [q - q^5 + O(q^6), q^2 - q^4 - q^5 + O(q^6), q^3 + O(q^6)]

    sage: S = CuspForms(Gamma1(3),6); S
    Cuspidal subspace of dimension 1 of Modular Forms space of dimension 3 for
    Congruence Subgroup Gamma1(3) of weight 6 over Rational Field
    sage: S.basis()
    [q - 6*q^2 + 9*q^3 + 4*q^4 + 6*q^5 + O(q^6)]
"""

#########################################################################
#       Copyright (C) 2004--2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  http://www.gnu.org/licenses/
#########################################################################

from sage.matrix.constructor import Matrix
from sage.matrix.special import identity_matrix
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_import import lazy_import
from sage.misc.verbose import verbose
from sage.rings.integer import Integer
from sage.rings.rational_field import QQ

lazy_import('sage.modular.modform.vm_basis', 'victor_miller_basis')

from .submodule import ModularFormsSubmodule
from . import weight1


class CuspidalSubmodule(ModularFormsSubmodule):
    """
    Base class for cuspidal submodules of ambient spaces of modular forms.
    """
    def __init__(self, ambient_space):
        """
        The cuspidal submodule of an ambient space of modular forms.

        EXAMPLES::

            sage: S = CuspForms(SL2Z,12); S
            Cuspidal subspace of dimension 1 of Modular Forms space of dimension 2 for
            Modular Group SL(2,Z) of weight 12 over Rational Field
            sage: S.basis()
            [q - 24*q^2 + 252*q^3 - 1472*q^4 + 4830*q^5 + O(q^6)]

            sage: S = CuspForms(Gamma0(33),2); S
            Cuspidal subspace of dimension 3 of Modular Forms space of dimension 6 for
            Congruence Subgroup Gamma0(33) of weight 2 over Rational Field
            sage: S.basis()
            [q - q^5 + O(q^6), q^2 - q^4 - q^5 + O(q^6), q^3 + O(q^6)]

            sage: S = CuspForms(Gamma1(3),6); S
            Cuspidal subspace of dimension 1 of Modular Forms space of dimension 3 for
            Congruence Subgroup Gamma1(3) of weight 6 over Rational Field
            sage: S.basis()
            [q - 6*q^2 + 9*q^3 + 4*q^4 + 6*q^5 + O(q^6)]
            sage: S == loads(dumps(S))
            True
        """
        from sage.misc.verbose import verbose
        verbose('creating cuspidal submodule of %s' % ambient_space)
        d = ambient_space._dim_cuspidal()
        V = ambient_space.module()
        G = [V.gen(i) for i in range(d)]
        S = V.submodule(G, check=False, already_echelonized=True)
        ModularFormsSubmodule.__init__(self, ambient_space, S)

    def _compute_q_expansion_basis(self, prec):
        r"""
        Compute a basis of `q`-expansions of ``self`` to the given precision. Not
        implemented in this abstract base class.

        EXAMPLES::

            sage: M = CuspForms(GammaH(11,[2]), 2)
            sage: sage.modular.modform.cuspidal_submodule.CuspidalSubmodule._compute_q_expansion_basis(M, 6)
            Traceback (most recent call last):
            ...
            NotImplementedError: q-expansion basis not implemented for "Cuspidal subspace of ..."
        """
        raise NotImplementedError('q-expansion basis not implemented for "%s"' % self)

    def _repr_(self) -> str:
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: S = CuspForms(Gamma1(3),6); S._repr_()
            'Cuspidal subspace of dimension 1 of Modular Forms space of dimension 3 for Congruence Subgroup Gamma1(3) of weight 6 over Rational Field'
        """
        return "Cuspidal subspace of dimension %s of %s" % (self.dimension(), self.ambient_module())

    def is_cuspidal(self) -> bool:
        """
        Return ``True`` since spaces of cusp forms are cuspidal.

        EXAMPLES::

            sage: CuspForms(4,10).is_cuspidal()
            True
        """
        return True

    @cached_method
    def modular_symbols(self, sign=0):
        """
        Return the corresponding space of modular symbols with the given sign.

        EXAMPLES::

            sage: S = ModularForms(11,2).cuspidal_submodule()
            sage: S.modular_symbols()
            Modular Symbols subspace of dimension 2 of Modular Symbols space
            of dimension 3 for Gamma_0(11) of weight 2 with sign 0 over Rational Field

            sage: S.modular_symbols(sign=-1)
            Modular Symbols subspace of dimension 1 of Modular Symbols space
            of dimension 1 for Gamma_0(11) of weight 2 with sign -1 over Rational Field

            sage: M = S.modular_symbols(sign=1); M
            Modular Symbols subspace of dimension 1 of Modular Symbols space of
            dimension 2 for Gamma_0(11) of weight 2 with sign 1 over Rational Field
            sage: M.sign()
            1

            sage: S = ModularForms(1,12).cuspidal_submodule()
            sage: S.modular_symbols()
            Modular Symbols subspace of dimension 2 of Modular Symbols space of
            dimension 3 for Gamma_0(1) of weight 12 with sign 0 over Rational Field

            sage: # needs sage.rings.number_field
            sage: eps = DirichletGroup(13).0
            sage: S = CuspForms(eps^2, 2)
            sage: S.modular_symbols(sign=0)
            Modular Symbols subspace of dimension 2 of Modular Symbols space
            of dimension 4 and level 13, weight 2, character [zeta6], sign 0,
            over Cyclotomic Field of order 6 and degree 2
            sage: S.modular_symbols(sign=1)
            Modular Symbols subspace of dimension 1 of Modular Symbols space
            of dimension 3 and level 13, weight 2, character [zeta6], sign 1,
            over Cyclotomic Field of order 6 and degree 2
            sage: S.modular_symbols(sign=-1)
            Modular Symbols subspace of dimension 1 of Modular Symbols space
            of dimension 1 and level 13, weight 2, character [zeta6], sign -1,
            over Cyclotomic Field of order 6 and degree 2
        """
        A = self.ambient_module()
        return A.modular_symbols(sign).cuspidal_submodule()

    def change_ring(self, R):
        r"""
        Change the base ring of ``self`` to ``R``, when this makes sense.

        This differs from
        :meth:`~sage.modular.modform.space.ModularFormsSpace.base_extend`
        in that there may not be a canonical map from ``self`` to the new
        space, as in the first example below. If this space has a
        character then this may fail when the character cannot be
        defined over ``R``, as in the second example.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: chi = DirichletGroup(109, CyclotomicField(3)).0
            sage: S9 = CuspForms(chi, 2, base_ring = CyclotomicField(9)); S9
            Cuspidal subspace of dimension 8 of
             Modular Forms space of dimension 10, character [zeta3 + 1] and weight 2
             over Cyclotomic Field of order 9 and degree 6
            sage: S9.change_ring(CyclotomicField(3))
            Cuspidal subspace of dimension 8 of
             Modular Forms space of dimension 10, character [zeta3 + 1] and weight 2
             over Cyclotomic Field of order 3 and degree 2
            sage: S9.change_ring(QQ)
            Traceback (most recent call last):
            ...
            ValueError: Space cannot be defined over Rational Field
        """
        return self.ambient_module().change_ring(R).cuspidal_submodule()


class CuspidalSubmodule_R(CuspidalSubmodule):
    """
    Cuspidal submodule over a non-minimal base ring.
    """
    def _compute_q_expansion_basis(self, prec):
        r"""
        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: CuspForms(Gamma1(13), 2, base_ring=QuadraticField(-7, 'a')).q_expansion_basis() # indirect doctest
            [q - 4*q^3 - q^4 + 3*q^5 + O(q^6), q^2 - 2*q^3 - q^4 + 2*q^5 + O(q^6)]
        """
        return ModularFormsSubmodule._compute_q_expansion_basis(self, prec)


class CuspidalSubmodule_modsym_qexp(CuspidalSubmodule):
    """
    Cuspidal submodule with `q`-expansions calculated via modular symbols.
    """
    def _compute_q_expansion_basis(self, prec=None):
        """
        Compute `q`-expansions of a basis for ``self`` (via modular symbols).

        EXAMPLES::

            sage: sage.modular.modform.cuspidal_submodule.CuspidalSubmodule_modsym_qexp(ModularForms(11,2))._compute_q_expansion_basis()
            [q - 2*q^2 - q^3 + 2*q^4 + q^5 + O(q^6)]
        """
        if prec is None:
            prec = self.prec()
        else:
            prec = Integer(prec)
        if self.dimension() == 0:
            return []
        M = self.modular_symbols(sign=1)
        return M.q_expansion_basis(prec)

    def _compute_hecke_matrix_prime(self, p):
        """
        Compute the matrix of a Hecke operator.

        EXAMPLES::

            sage: C = CuspForms(38, 2)
            sage: C._compute_hecke_matrix_prime(7)
            [-1  0  0  0]
            [ 0 -1  0  0]
            [-2 -2  1 -2]
            [ 2  2 -2  1]
        """
        A = self.modular_symbols(sign=1)
        T = A.hecke_matrix(p)
        return _convert_matrix_from_modsyms(A, T)[0]

    def hecke_polynomial(self, n, var='x'):
        r"""
        Return the characteristic polynomial of the Hecke operator `T_n` on this
        space. This is computed via modular symbols, and in particular is
        faster to compute than the matrix itself.

        EXAMPLES::

            sage: CuspForms(105, 2).hecke_polynomial(2, 'y')
            y^13 + 5*y^12 - 4*y^11 - 52*y^10 - 34*y^9 + 174*y^8 + 212*y^7
             - 196*y^6 - 375*y^5 - 11*y^4 + 200*y^3 + 80*y^2

        Check that this gives the same answer as computing the Hecke matrix::

            sage: CuspForms(105, 2).hecke_matrix(2).charpoly(var='y')
            y^13 + 5*y^12 - 4*y^11 - 52*y^10 - 34*y^9 + 174*y^8 + 212*y^7
             - 196*y^6 - 375*y^5 - 11*y^4 + 200*y^3 + 80*y^2

        Check that :issue:`21546` is fixed (this example used to take about 5 hours)::

            sage: CuspForms(1728, 2).hecke_polynomial(2) # long time (20 sec)
            x^253 + x^251 - 2*x^249
        """
        return self.modular_symbols(sign=1).hecke_polynomial(n, var)

    def new_submodule(self, p=None):
        r"""
        Return the new subspace of this space of cusp forms. This is computed
        using modular symbols.

        EXAMPLES::

            sage: CuspForms(55).new_submodule()
            Modular Forms subspace of dimension 3 of
             Modular Forms space of dimension 8 for
              Congruence Subgroup Gamma0(55) of weight 2 over Rational Field
        """
        symbs = self.modular_symbols(sign=1).new_subspace(p)
        bas = []
        for x in symbs.q_expansion_basis(self.sturm_bound()):
            bas.append(self(x))
        return self.submodule(bas, check=False)


class CuspidalSubmodule_level1_Q(CuspidalSubmodule):
    r"""
    Space of cusp forms of level 1 over `\QQ`.
    """
    def _compute_q_expansion_basis(self, prec=None):
        """
        Compute `q`-expansions of a basis for ``self``.

        EXAMPLES::

            sage: sage.modular.modform.cuspidal_submodule.CuspidalSubmodule_level1_Q(ModularForms(1,12))._compute_q_expansion_basis()
            [q - 24*q^2 + 252*q^3 - 1472*q^4 + 4830*q^5 + O(q^6)]
        """
        if prec is None:
            prec = self.prec()
        else:
            prec = Integer(prec)
        return victor_miller_basis(self.weight(), prec, cusp_only=True, var='q')

    def _pari_init_(self):
        """
        Conversion to Pari.

        EXAMPLES::

            sage: A = ModularForms(1,12).cuspidal_submodule()
            sage: pari.mfparams(A)
            [1, 12, 1, 1, t - 1]
            sage: pari.mfdim(A)
            1
        """
        from sage.libs.pari import pari
        return pari.mfinit([self.level(), self.weight()], 1)


class CuspidalSubmodule_wt1_eps(CuspidalSubmodule):
    r"""
    Space of cusp forms of weight 1 with specified character.
    """

    def _compute_q_expansion_basis(self, prec=None):
        r"""
        Compute `q`-expansion basis using Schaeffer's algorithm.

        EXAMPLES::

            sage: CuspForms(DirichletGroup(23, QQ).0, 1).q_echelon_basis()  # indirect doctest
            [q - q^2 - q^3 + O(q^6)]
        """
        if prec is None:
            prec = self.prec()
        else:
            prec = Integer(prec)
        chi = self.character()
        return [weight1.modular_ratio_to_prec(chi, f, prec) for f in
            weight1.hecke_stable_subspace(chi)]

    def _pari_init_(self):
        """
        Conversion to Pari.

        EXAMPLES::

            sage: A = CuspForms(DirichletGroup(23, QQ).0, 1)
            sage: pari.mfparams(A)
            [23, 1, -23, 1, t + 1]
            sage: pari.mfdim(A)
            1
        """
        from sage.libs.pari import pari
        return pari.mfinit([self.level(), self.weight(), self.character()], 1)


class CuspidalSubmodule_wt1_gH(CuspidalSubmodule):
    r"""
    Space of cusp forms of weight 1 for a GammaH group.
    """

    def _compute_q_expansion_basis(self, prec=None):
        r"""
        Compute `q`-expansion basis using Schaeffer's algorithm.

        EXAMPLES::

            sage: CuspForms(GammaH(31, [7]), 1).q_expansion_basis()  # indirect doctest
            [q - q^2 - q^5 + O(q^6)]

        A more elaborate example (two Galois-conjugate characters each giving a
        2-dimensional space)::

            sage: CuspForms(GammaH(124, [85]), 1).q_expansion_basis()  # long time
            [q - q^4 - q^6 + O(q^7), q^2 + O(q^7), q^3 + O(q^7), q^5 - q^6 + O(q^7)]
        """
        if prec is None:
            prec = self.prec()
        else:
            prec = Integer(prec)

        chars = self.group().characters_mod_H(sign=-1, galois_orbits=True)

        B = []
        dim = 0
        for c in chars:
            chi = c.minimize_base_ring()
            Bchi = [weight1.modular_ratio_to_prec(chi, f, prec)
                for f in weight1.hecke_stable_subspace(chi) ]
            if Bchi == []:
                continue
            if chi.base_ring() == QQ:
                B += [f.padded_list(prec) for f in Bchi]
                dim += len(Bchi)
            else:
                d = chi.base_ring().degree()
                dim += d * len(Bchi)
                for f in Bchi:
                    w = f.padded_list(prec)
                    for i in range(d):
                        B.append([x[i] for x in w])

        basis_mat = Matrix(QQ, B)
        if basis_mat.is_zero():
            return []
        # Daft thing: "transformation=True" parameter to echelonize
        # is ignored for rational matrices!
        big_mat = basis_mat.augment(identity_matrix(dim))
        big_mat.echelonize()
        c = big_mat.pivots()[-1]

        echelon_basis_mat = big_mat[:, :prec]

        R = self._q_expansion_ring()

        if c >= prec:
            verbose("Precision %s insufficient to determine basis" % prec, level=1)
        else:
            verbose("Minimal precision for basis: %s" % (c+1), level=1)
            t = big_mat[:, prec:]
            assert echelon_basis_mat == t * basis_mat
            self.__transformation_matrix = t
            self._char_basis = [R(f.list(), c+1) for f in basis_mat.rows()]

        return [R(f.list(), prec) for f in echelon_basis_mat.rows() if f != 0]

    def _transformation_matrix(self):
        r"""
        Return the transformation matrix between a basis of Galois orbits of
        forms with character, and the echelon-form basis.

        EXAMPLES::

            sage: CuspForms(GammaH(31, [7]), 1)._transformation_matrix()
            [1]
            sage: CuspForms(GammaH(124, [33]), 1)._transformation_matrix()  # long time
            [ 1  1  0  0  0  0  1]
            [ 0  0  0  0  0  1  0]
            [ 1  0  1  1 -1 -1  1]
            [ 0  0  0  0  0  0  1]
            [ 0  1  0  0  0  0  0]
            [ 1  0  0  0 -1  0  1]
            [ 0  0  1  0  0 -1  0]
        """
        self.q_expansion_basis() # triggers iterative computation
        return self.__transformation_matrix

    def _compute_diamond_matrix(self, d):
        r"""
        EXAMPLES::

            sage: CuspForms(GammaH(31, [7]), 1).diamond_bracket_matrix(3)
            [-1]

            sage: C = CuspForms(GammaH(124, [33]), 1)   # long time
            sage: D = C.diamond_bracket_matrix(3); D    # long time
            [ 0  0  0 -1  1  0  0]
            [ 0 -1  0  0  0  0  0]
            [ 2  1  1 -2 -1 -2 -1]
            [ 0  0  0 -1  0  0  0]
            [-1  0  0  1  1  0  0]
            [ 2  0  0 -2 -1 -1  0]
            [ 0  2  1  0  0 -1  0]
            sage: t = C._transformation_matrix(); ~t * D * t   # long time
            [ 1  1  0  0  0  0  0]
            [-1  0  0  0  0  0  0]
            [ 0  0  1  1  0  0  0]
            [ 0  0 -1  0  0  0  0]
            [ 0  0  0  0 -1  0  0]
            [ 0  0  0  0  0 -1  0]
            [ 0  0  0  0  0  0 -1]
        """
        chars = self.group().characters_mod_H(sign=-1, galois_orbits=True)
        A = Matrix(QQ, 0, 0)
        for c in chars:
            chi = c.minimize_base_ring()
            dim = weight1.dimension_wt1_cusp_forms(chi)
            if chi.base_ring() == QQ:
                m = Matrix(QQ, 1, 1, [chi(d)])
            else:
                m = chi(d).matrix().transpose()
            for i in range(dim):
                A = A.block_sum(m)
        t = self._transformation_matrix()
        return t * A * ~t

    def _compute_hecke_matrix(self, n):
        r"""
        EXAMPLES::

            sage: CuspForms(GammaH(31, [7]), 1).hecke_matrix(7)
            [-1]

            sage: # long time
            sage: C = CuspForms(GammaH(124, [33]), 1)
            sage: C.hecke_matrix(2)
            [ 0  0 -1 -1  0  1  0]
            [ 1  0  0 -1 -1 -1  0]
            [ 0  0  0 -1  1  1 -1]
            [ 0  1  0 -1  0  0  0]
            [ 0  0 -1  0  0  1  1]
            [ 0  0  0 -1  0  0 -1]
            [ 0  0  0  0  0  1  0]
            sage: C.hecke_matrix(7)
            [ 0  1  0 -1  0  0  1]
            [ 0 -1  0  0  0  0  0]
            [ 0  1 -1  0  0  0  1]
            [ 0  0  0 -1  0  0  0]
            [ 0  1  1  0  0 -1  0]
            [ 1  0 -1 -1 -1  0  1]
            [ 0  1  0  0  1  0  0]
            sage: C.hecke_matrix(23) == 0
            True
        """
        chars = self.group().characters_mod_H(sign=-1, galois_orbits=True)
        A = Matrix(QQ, 0, 0)
        for c in chars:
            chi = c.minimize_base_ring()
            d = weight1.dimension_wt1_cusp_forms(chi)
            e = chi.base_ring().degree()
            H = Matrix(QQ, d*e, d*e)
            from .constructor import CuspForms
            M = CuspForms(chi, 1).hecke_matrix(n)
            if e == 1:
                H = M
            else:
                for i in range(d):
                    for j in range(d):
                        H[e*i: e*(i+1), e*j: e*(j+1)] = M[i, j].matrix().transpose()
            A = A.block_sum(H)
        t = self._transformation_matrix()
        return t * A * ~t


class CuspidalSubmodule_g0_Q(CuspidalSubmodule_modsym_qexp):
    r"""
    Space of cusp forms for `\Gamma_0(N)` over `\QQ`.
    """
    def _pari_init_(self):
        """
        Conversion to Pari.

        EXAMPLES::

            sage: h = Newforms(37)[1]
            sage: MF = h.parent()
            sage: pari.mfparams(MF)
            [37, 2, 1, 1, t - 1]
            sage: pari.mfdim(MF)
            2
        """
        from sage.libs.pari import pari
        return pari.mfinit([self.level(), self.weight()], 1)


class CuspidalSubmodule_gH_Q(CuspidalSubmodule_modsym_qexp):
    r"""
    Space of cusp forms for `\Gamma_H(N)` over `\QQ`.
    """

    def _compute_hecke_matrix(self, n):
        r"""
        Compute the matrix of the Hecke operator `T_n` acting on this space.
        This is done directly using modular symbols, rather than using
        `q`-expansions as for spaces with fixed character.

        EXAMPLES::

            sage: CuspForms(Gamma1(8), 4)._compute_hecke_matrix(2)
            [  0 -16  32]
            [  1 -10  18]
            [  0  -4   8]

            sage: CuspForms(GammaH(15, [4]), 3)._compute_hecke_matrix(17)
            [ 18  22 -30 -60]
            [  4   0   6 -18]
            [  6   0   2 -20]
            [  6  12 -24 -20]
        """
        symbs = self.modular_symbols(sign=1)
        return _convert_matrix_from_modsyms(symbs, symbs.hecke_matrix(n))[0]

    def _compute_diamond_matrix(self, d):
        r"""
        EXAMPLES::

            sage: CuspForms(Gamma1(5), 6).diamond_bracket_matrix(3)  # indirect doctest
            [ -1   0   0]
            [  3   5 -12]
            [  1   2  -5]
            sage: CuspForms(GammaH(15, [4]), 3)._compute_diamond_matrix(7)
            [ 2  3 -9 -3]
            [ 0  2 -3  0]
            [ 0  1 -2  0]
            [ 1  1 -3 -2]
        """
        symbs = self.modular_symbols(sign=1)
        return _convert_matrix_from_modsyms(symbs, symbs.diamond_bracket_matrix(d))[0]


class CuspidalSubmodule_g1_Q(CuspidalSubmodule_gH_Q):
    r"""
    Space of cusp forms for `\Gamma_1(N)` over `\QQ`.
    """


class CuspidalSubmodule_eps(CuspidalSubmodule_modsym_qexp):
    """
    Space of cusp forms with given Dirichlet character.

    EXAMPLES::

        sage: S = CuspForms(DirichletGroup(5).0,5); S
        Cuspidal subspace of dimension 1 of Modular Forms space of dimension 3,
        character [zeta4] and weight 5 over Cyclotomic Field of order 4 and degree 2

        sage: S.basis()
        [q + (-zeta4 - 1)*q^2 + (6*zeta4 - 6)*q^3 - 14*zeta4*q^4 + (15*zeta4 + 20)*q^5 + O(q^6)]
        sage: f = S.0
        sage: f.qexp()
        q + (-zeta4 - 1)*q^2 + (6*zeta4 - 6)*q^3 - 14*zeta4*q^4 + (15*zeta4 + 20)*q^5 + O(q^6)
        sage: f.qexp(7)
        q + (-zeta4 - 1)*q^2 + (6*zeta4 - 6)*q^3 - 14*zeta4*q^4 + (15*zeta4 + 20)*q^5 + 12*q^6 + O(q^7)
        sage: f.qexp(3)
        q + (-zeta4 - 1)*q^2 + O(q^3)
        sage: f.qexp(2)
        q + O(q^2)
        sage: f.qexp(1)
        O(q^1)
    """
    pass


def _convert_matrix_from_modsyms(symbs, T):
    r"""
    Given a space of modular symbols and a matrix T acting on it, calculate the
    matrix of the corresponding operator on the echelon-form basis of the
    corresponding space of modular forms.

    The matrix T *must* commute with the Hecke operators! We use this when T is
    either a Hecke operator, or a diamond operator. This will *not work* for
    the Atkin-Lehner operators, for instance, when there are oldforms present.

    OUTPUT:

    A pair `(T_e, ps)` with `T_e` the converted matrix and `ps` a list
    of pivot elements of the echelon basis.

    EXAMPLES::

        sage: CuspForms(Gamma1(5), 6).diamond_bracket_matrix(3)  # indirect doctest
        [ -1   0   0]
        [  3   5 -12]
        [  1   2  -5]
    """
    d = symbs.rank()

    # create a vector space of appropriate dimension to
    # contain our q-expansions
    A = symbs.base_ring()
    r = symbs.sturm_bound()
    X = A**r
    Y = X.zero_submodule()
    basis = []
    basis_images = []

    # we repeatedly use these matrices below, so we store them
    # once as lists to save time.
    hecke_matrix_ls = [symbs.hecke_matrix(m).list()
                       for m in range(1, r + 1)]
    hecke_image_ls = [(T * symbs.hecke_matrix(m)).list()
                      for m in range(1, r + 1)]

    # compute the q-expansions of some cusp forms and their
    # images under T_n
    for i in range(d**2):
        v = X([hecke_matrix_ls[m][i] for m in range(r)])
        Ynew = Y.span(Y.basis() + [v])
        if Ynew.rank() > Y.rank():
            basis.append(v)
            basis_images.append(X([hecke_image_ls[m][i] for m in range(r)]))
            Y = Ynew
            if len(basis) == d:
                break

    # now we can compute the matrix acting on the echelonized space of mod forms
    # need to pass A as base ring since otherwise there are problems when the
    # space has dimension 0
    bigmat = Matrix(A, basis).augment(Matrix(A, basis_images))
    bigmat.echelonize()
    pivs = bigmat.pivots()
    return bigmat.matrix_from_rows_and_columns(list(range(d)),
                                               [r + x for x in pivs]), pivs



# distutils: libraries = flint
# distutils: depends = flint/fq.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void fq_ctx_init(fq_ctx_t ctx, const fmpz_t p, slong d, const char * var) noexcept
    int _fq_ctx_init_conway(fq_ctx_t ctx, const fmpz_t p, slong d, const char * var) noexcept
    void fq_ctx_init_conway(fq_ctx_t ctx, const fmpz_t p, slong d, const char * var) noexcept
    void fq_ctx_init_modulus(fq_ctx_t ctx, const fmpz_mod_poly_t modulus, const fmpz_mod_ctx_t ctxp, const char * var) noexcept
    void fq_ctx_clear(fq_ctx_t ctx) noexcept
    const fmpz_mod_poly_struct* fq_ctx_modulus(const fq_ctx_t ctx) noexcept
    slong fq_ctx_degree(const fq_ctx_t ctx) noexcept
    const fmpz * fq_ctx_prime(const fq_ctx_t ctx) noexcept
    void fq_ctx_order(fmpz_t f, const fq_ctx_t ctx) noexcept
    int fq_ctx_fprint(FILE * file, const fq_ctx_t ctx) noexcept
    void fq_ctx_print(const fq_ctx_t ctx) noexcept
    void fq_ctx_randtest(fq_ctx_t ctx, flint_rand_t state) noexcept
    void fq_ctx_randtest_reducible(fq_ctx_t ctx, flint_rand_t state) noexcept
    void fq_init(fq_t rop, const fq_ctx_t ctx) noexcept
    void fq_init2(fq_t rop, const fq_ctx_t ctx) noexcept
    void fq_clear(fq_t rop, const fq_ctx_t ctx) noexcept
    void _fq_sparse_reduce(fmpz * R, slong lenR, const fq_ctx_t ctx) noexcept
    void _fq_dense_reduce(fmpz * R, slong lenR, const fq_ctx_t ctx) noexcept
    void _fq_reduce(fmpz * r, slong lenR, const fq_ctx_t ctx) noexcept
    void fq_reduce(fq_t rop, const fq_ctx_t ctx) noexcept
    void fq_add(fq_t rop, const fq_t op1, const fq_t op2, const fq_ctx_t ctx) noexcept
    void fq_sub(fq_t rop, const fq_t op1, const fq_t op2, const fq_ctx_t ctx) noexcept
    void fq_sub_one(fq_t rop, const fq_t op1, const fq_ctx_t ctx) noexcept
    void fq_neg(fq_t rop, const fq_t op, const fq_ctx_t ctx) noexcept
    void fq_mul(fq_t rop, const fq_t op1, const fq_t op2, const fq_ctx_t ctx) noexcept
    void fq_mul_fmpz(fq_t rop, const fq_t op, const fmpz_t x, const fq_ctx_t ctx) noexcept
    void fq_mul_si(fq_t rop, const fq_t op, slong x, const fq_ctx_t ctx) noexcept
    void fq_mul_ui(fq_t rop, const fq_t op, ulong x, const fq_ctx_t ctx) noexcept
    void fq_sqr(fq_t rop, const fq_t op, const fq_ctx_t ctx) noexcept
    void fq_div(fq_t rop, const fq_t op1, const fq_t op2, const fq_ctx_t ctx) noexcept
    void _fq_inv(fmpz * rop, const fmpz * op, slong len, const fq_ctx_t ctx) noexcept
    void fq_inv(fq_t rop, const fq_t op, const fq_ctx_t ctx) noexcept
    void fq_gcdinv(fq_t f, fq_t inv, const fq_t op, const fq_ctx_t ctx) noexcept
    void _fq_pow(fmpz * rop, const fmpz * op, slong len, const fmpz_t e, const fq_ctx_t ctx) noexcept
    void fq_pow(fq_t rop, const fq_t op, const fmpz_t e, const fq_ctx_t ctx) noexcept
    void fq_pow_ui(fq_t rop, const fq_t op, const ulong e, const fq_ctx_t ctx) noexcept
    int fq_sqrt(fq_t rop, const fq_t op1, const fq_ctx_t ctx) noexcept
    void fq_pth_root(fq_t rop, const fq_t op1, const fq_ctx_t ctx) noexcept
    bint fq_is_square(const fq_t op, const fq_ctx_t ctx) noexcept
    int fq_fprint_pretty(FILE * file, const fq_t op, const fq_ctx_t ctx) noexcept
    int fq_print_pretty(const fq_t op, const fq_ctx_t ctx) noexcept
    int fq_fprint(FILE * file, const fq_t op, const fq_ctx_t ctx) noexcept
    void fq_print(const fq_t op, const fq_ctx_t ctx) noexcept
    char * fq_get_str(const fq_t op, const fq_ctx_t ctx) noexcept
    char * fq_get_str_pretty(const fq_t op, const fq_ctx_t ctx) noexcept
    void fq_randtest(fq_t rop, flint_rand_t state, const fq_ctx_t ctx) noexcept
    void fq_randtest_not_zero(fq_t rop, flint_rand_t state, const fq_ctx_t ctx) noexcept
    void fq_randtest_dense(fq_t rop, flint_rand_t state, const fq_ctx_t ctx) noexcept
    void fq_rand(fq_t rop, flint_rand_t state, const fq_ctx_t ctx) noexcept
    void fq_rand_not_zero(fq_t rop, flint_rand_t state, const fq_ctx_t ctx) noexcept
    void fq_set(fq_t rop, const fq_t op, const fq_ctx_t ctx) noexcept
    void fq_set_si(fq_t rop, const slong x, const fq_ctx_t ctx) noexcept
    void fq_set_ui(fq_t rop, const ulong x, const fq_ctx_t ctx) noexcept
    void fq_set_fmpz(fq_t rop, const fmpz_t x, const fq_ctx_t ctx) noexcept
    void fq_swap(fq_t op1, fq_t op2, const fq_ctx_t ctx) noexcept
    void fq_zero(fq_t rop, const fq_ctx_t ctx) noexcept
    void fq_one(fq_t rop, const fq_ctx_t ctx) noexcept
    void fq_gen(fq_t rop, const fq_ctx_t ctx) noexcept
    int fq_get_fmpz(fmpz_t rop, const fq_t op, const fq_ctx_t ctx) noexcept
    void fq_get_fmpz_poly(fmpz_poly_t a, const fq_t b, const fq_ctx_t ctx) noexcept
    void fq_get_fmpz_mod_poly(fmpz_mod_poly_t a, const fq_t b, const fq_ctx_t ctx) noexcept
    void fq_set_fmpz_poly(fq_t a, const fmpz_poly_t b, const fq_ctx_t ctx) noexcept
    void fq_set_fmpz_mod_poly(fq_t a, const fmpz_mod_poly_t b, const fq_ctx_t ctx) noexcept
    void fq_get_fmpz_mod_mat(fmpz_mod_mat_t col, const fq_t a, const fq_ctx_t ctx) noexcept
    void fq_set_fmpz_mod_mat(fq_t a, const fmpz_mod_mat_t col, const fq_ctx_t ctx) noexcept
    bint fq_is_zero(const fq_t op, const fq_ctx_t ctx) noexcept
    bint fq_is_one(const fq_t op, const fq_ctx_t ctx) noexcept
    bint fq_equal(const fq_t op1, const fq_t op2, const fq_ctx_t ctx) noexcept
    bint fq_is_invertible(const fq_t op, const fq_ctx_t ctx) noexcept
    bint fq_is_invertible_f(fq_t f, const fq_t op, const fq_ctx_t ctx) noexcept
    void _fq_trace(fmpz_t rop, const fmpz * op, slong len, const fq_ctx_t ctx) noexcept
    void fq_trace(fmpz_t rop, const fq_t op, const fq_ctx_t ctx) noexcept
    void _fq_norm(fmpz_t rop, const fmpz * op, slong len, const fq_ctx_t ctx) noexcept
    void fq_norm(fmpz_t rop, const fq_t op, const fq_ctx_t ctx) noexcept
    void _fq_frobenius(fmpz * rop, const fmpz * op, slong len, slong e, const fq_ctx_t ctx) noexcept
    void fq_frobenius(fq_t rop, const fq_t op, slong e, const fq_ctx_t ctx) noexcept
    int fq_multiplicative_order(fmpz * ord, const fq_t op, const fq_ctx_t ctx) noexcept
    bint fq_is_primitive(const fq_t op, const fq_ctx_t ctx) noexcept
    void fq_bit_pack(fmpz_t f, const fq_t op, flint_bitcnt_t bit_size, const fq_ctx_t ctx) noexcept
    void fq_bit_unpack(fq_t rop, const fmpz_t f, flint_bitcnt_t bit_size, const fq_ctx_t ctx) noexcept



from sage.structure.element cimport MultiplicativeGroupElement, Element, MonoidElement, Matrix

cpdef is_MatrixGroupElement(x)

cdef class MatrixGroupElement_generic(MultiplicativeGroupElement):
    cdef public Matrix _matrix

    cpdef _act_on_(self, x, bint self_on_left)
    cpdef _mul_(self, other)
    cpdef list list(self)



"""
all.py -- export of abelian groups to Sage
"""

# *****************************************************************************
#
#   Sage: Open Source Mathematical Software
#
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

#from dual_abelian_group import DualAbelianGroup
from sage.groups.abelian_gps.abelian_group import AbelianGroup, word_problem
from sage.groups.abelian_gps.values import AbelianGroupWithValues

# TODO:
# Implement group homset, conversion of generator images to morphism
from sage.groups.abelian_gps.abelian_group_morphism import AbelianGroupMorphism



# sage.doctest: needs sage.symbolic
"""
Density plots
"""

#*****************************************************************************
#       Copyright (C) 2006 Alex Clemesha <clemesha@gmail.com>,
#                          William Stein <wstein@gmail.com>,
#                     2008 Mike Hansen <mhansen@gmail.com>,
#                          Arnaud Bergeron <abergeron@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************
from sage.plot.primitive import GraphicPrimitive
from sage.misc.decorators import options
from sage.plot.colors import get_cmap
from sage.arith.srange import xsrange


class DensityPlot(GraphicPrimitive):
    """
    Primitive class for the density plot graphics type.  See
    ``density_plot?`` for help actually doing density plots.

    INPUT:

    - ``xy_data_array`` -- list of lists giving evaluated values of the
      function on the grid

    - ``xrange`` -- tuple of 2 floats indicating range for horizontal direction

    - ``yrange`` -- tuple of 2 floats indicating range for vertical direction

    - ``options`` -- dictionary of valid plot options to pass to constructor

    EXAMPLES:

    Note this should normally be used indirectly via ``density_plot``::

        sage: from sage.plot.density_plot import DensityPlot
        sage: D = DensityPlot([[1,3],[2,4]], (1,2), (2,3),options={})
        sage: D
        DensityPlot defined by a 2 x 2 data grid
        sage: D.yrange
        (2, 3)
        sage: D.options()
        {}

    TESTS:

    We test creating a density plot::

        sage: x,y = var('x,y')
        sage: density_plot(x^2 - y^3 + 10*sin(x*y), (x,-4,4), (y,-4,4), plot_points=121, cmap='hsv')
        Graphics object consisting of 1 graphics primitive
    """
    def __init__(self, xy_data_array, xrange, yrange, options):
        """
        Initialize base class ``DensityPlot``.

        EXAMPLES::

            sage: x,y = var('x,y')
            sage: D = density_plot(x^2 - y^3 + 10*sin(x*y), (x,-4,4), (y,-4,4), plot_points=121, cmap='hsv')
            sage: D[0].xrange
            (-4.0, 4.0)
            sage: D[0].options()['plot_points']
            121
        """
        self.xrange = xrange
        self.yrange = yrange
        self.xy_data_array = xy_data_array
        self.xy_array_row = len(xy_data_array)
        self.xy_array_col = len(xy_data_array[0])
        GraphicPrimitive.__init__(self, options)

    def get_minmax_data(self):
        """
        Return a dictionary with the bounding box data.

        EXAMPLES::

            sage: x,y = var('x,y')
            sage: f(x, y) = x^2 + y^2
            sage: d = density_plot(f, (3,6), (3,6))[0].get_minmax_data()
            sage: d['xmin']
            3.0
            sage: d['ymin']
            3.0
        """
        from sage.plot.plot import minmax_data
        return minmax_data(self.xrange, self.yrange, dict=True)

    def _allowed_options(self):
        """
        Return the allowed options for the DensityPlot class.

        TESTS::

            sage: isinstance(density_plot(x, (-2,3), (1,10))[0]._allowed_options(), dict)
            True
        """
        return {'plot_points': 'How many points to use for plotting precision',
                'cmap': """the name of a predefined colormap,
                       a list of colors or an instance of a
                       matplotlib Colormap. Type: import matplotlib.cm; matplotlib.cm.datad.keys()
                       for available colormap names.""",
                'interpolation': 'What interpolation method to use'}

    def _repr_(self):
        """
        String representation of DensityrPlot primitive.

        EXAMPLES::

            sage: x,y = var('x,y')
            sage: D = density_plot(x^2 - y^2, (x,-2,2), (y,-2,2))
            sage: d = D[0]; d
            DensityPlot defined by a 25 x 25 data grid
        """
        return "DensityPlot defined by a {} x {} data grid".format(self.xy_array_row, self.xy_array_col)

    def _render_on_subplot(self, subplot):
        """
        TESTS:

        A somewhat random plot, but fun to look at::

            sage: x,y = var('x,y')
            sage: density_plot(x^2 - y^3 + 10*sin(x*y), (x,-4,4), (y,-4,4), plot_points=121, cmap='hsv')
            Graphics object consisting of 1 graphics primitive
        """
        options = self.options()
        cmap = get_cmap(options['cmap'])

        x0, x1 = float(self.xrange[0]), float(self.xrange[1])
        y0, y1 = float(self.yrange[0]), float(self.yrange[1])

        subplot.imshow(self.xy_data_array, origin='lower',
                       cmap=cmap, extent=(x0,x1,y0,y1),
                       interpolation=options['interpolation'])


@options(plot_points=25, cmap='gray', interpolation='catrom')
def density_plot(f, xrange, yrange, **options):
    r"""
    ``density_plot`` takes a function of two variables, `f(x,y)`
    and plots the height of the function over the specified
    ``xrange`` and ``yrange`` as demonstrated below.

    ``density_plot(f, (xmin,xmax), (ymin,ymax), ...)``

    INPUT:

    - ``f`` -- a function of two variables

    - ``(xmin, xmax)`` -- 2-tuple, the range of ``x`` values OR 3-tuple
      ``(x,xmin,xmax)``

    - ``(ymin, ymax)`` -- 2-tuple, the range of ``y`` values OR 3-tuple
      ``(y,ymin,ymax)``

    The following inputs must all be passed in as named parameters:

    - ``plot_points`` -- integer (default: 25); number of points to plot
      in each direction of the grid

    - ``cmap`` -- a colormap (default: ``'gray'``), the name of
      a predefined colormap, a list of colors or an instance of a matplotlib
      Colormap. Type: ``import matplotlib.cm; matplotlib.cm.datad.keys()``
      for available colormap names.

    - ``interpolation`` -- string (default: ``'catrom'``); the interpolation
      method to use: ``'bilinear'``, ``'bicubic'``, ``'spline16'``,
      ``'spline36'``, ``'quadric'``, ``'gaussian'``, ``'sinc'``,
      ``'bessel'``, ``'mitchell'``, ``'lanczos'``, ``'catrom'``,
      ``'hermite'``, ``'hanning'``, ``'hamming'``, ``'kaiser'``

    EXAMPLES:

    Here we plot a simple function of two variables.  Note that
    since the input function is an expression, we need to explicitly
    declare the variables in 3-tuples for the range::

        sage: x,y = var('x,y')
        sage: density_plot(sin(x) * sin(y), (x,-2,2), (y,-2,2))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        x,y = var('x,y')
        g = density_plot(sin(x) * sin(y), (x,-2,2), (y,-2,2))
        sphinx_plot(g)

    Here we change the ranges and add some options; note that here
    ``f`` is callable (has variables declared), so we can use 2-tuple ranges::

        sage: x,y = var('x,y')
        sage: f(x,y) = x^2 * cos(x*y)
        sage: density_plot(f, (x,-10,5), (y,-5,5), interpolation='sinc', plot_points=100)
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        x,y = var('x,y')
        def f(x, y): return x**2 * cos(x*y)
        g = density_plot(f, (x,-10,5), (y,-5,5), interpolation='sinc', plot_points=100)
        sphinx_plot(g)

    An even more complicated plot::

        sage: x,y = var('x,y')
        sage: density_plot(sin(x^2+y^2) * cos(x) * sin(y), (x,-4,4), (y,-4,4), cmap='jet', plot_points=100)
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        x,y = var('x,y')
        g = density_plot(sin(x**2 + y**2)*cos(x)*sin(y), (x,-4,4), (y,-4,4), cmap='jet', plot_points=100)
        sphinx_plot(g)

    This should show a "spotlight" right on the origin::

        sage: x,y = var('x,y')
        sage: density_plot(1/(x^10 + y^10), (x,-10,10), (y,-10,10))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        x,y = var('x,y')
        g = density_plot(1/(x**10 + y**10), (x,-10,10), (y,-10,10))
        sphinx_plot(g)

    Some elliptic curves, but with symbolic endpoints.  In the first
    example, the plot is rotated 90 degrees because we switch the
    variables `x`, `y`::

        sage: density_plot(y^2 + 1 - x^3 - x, (y,-pi,pi), (x,-pi,pi))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        x,y = var('x,y')
        g = density_plot(y**2 + 1 - x**3 - x, (y,-pi,pi), (x,-pi,pi))
        sphinx_plot(g)

    ::

        sage: density_plot(y^2 + 1 - x^3 - x, (x,-pi,pi), (y,-pi,pi))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        x,y = var('x,y')
        g = density_plot(y**2 + 1 - x**3 - x, (x,-pi,pi), (y,-pi,pi))
        sphinx_plot(g)

    Extra options will get passed on to show(), as long as they are valid::

        sage: density_plot(log(x) + log(y), (x,1,10), (y,1,10), aspect_ratio=1)
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        x,y = var('x,y')
        g = density_plot(log(x) + log(y), (x,1,10), (y,1,10), aspect_ratio=1)
        sphinx_plot(g)

    ::

        sage: density_plot(log(x) + log(y), (x,1,10), (y,1,10)).show(aspect_ratio=1) # These are equivalent

    TESTS:

    Check that :issue:`15315` is fixed, i.e., density_plot respects the
    ``aspect_ratio`` parameter. Without the fix, it looks like a thin line
    of width a few mm. With the fix it should look like a nice fat layered
    image::

        sage: density_plot((x*y)^(1/2), (x,0,3), (y,0,500), aspect_ratio=.01)
        Graphics object consisting of 1 graphics primitive

    Default ``aspect_ratio`` is ``'automatic'``, and that should work too::

        sage: density_plot((x*y)^(1/2), (x,0,3), (y,0,500))
        Graphics object consisting of 1 graphics primitive

    Check that :issue:`17684` is fixed, i.e., symbolic values can be plotted::

        sage: def f(x, y):
        ....:     return SR(x)
        sage: density_plot(f, (0,1), (0,1))
        Graphics object consisting of 1 graphics primitive
    """
    from sage.plot.graphics import Graphics
    from sage.plot.misc import setup_for_eval_on_grid
    from sage.rings.real_double import RDF
    g, ranges = setup_for_eval_on_grid([f], [xrange, yrange], options['plot_points'])
    g = g[0]
    xrange, yrange = (r[:2] for r in ranges)

    xy_data_array = [[RDF(g(x,y)) for x in xsrange(*ranges[0], include_endpoint=True)]
                            for y in xsrange(*ranges[1], include_endpoint=True)]

    g = Graphics()
    g._set_extra_kwds(Graphics._extract_kwds_for_show(options, ignore=['xmin','xmax']))
    g.add_primitive(DensityPlot(xy_data_array, xrange, yrange, options))
    return g



# ****************************************************************************
#       Copyright (C) 2021 Xavier Caruso <xavier.caruso@normalesup.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


from sage.rings.padics.precision_error import PrecisionError


cdef inline int ERROR_ABANDON = 1
cdef inline int ERROR_NOTDEFINED = 1 << 1
cdef inline int ERROR_PRECISION = 1 << 2
cdef inline int ERROR_OVERFLOW = 1 << 3
cdef inline int ERROR_NOTSQUARE = 1 << 4  # maybe we should have something more generic here
cdef inline int ERROR_INTEGRAL = 1 << 5
cdef inline int ERROR_DIVISION = 1 << 6
cdef inline int ERROR_CIRCULAR = 1 << 7

cdef inline int ERROR_UNEXPECTED = 1 << 30


def raise_error(error, permissive=False):
    r"""
    Raise an error according to the given error code.

    INPUT:

    - ``error`` -- integer; the error code

    - ``permissive`` -- boolean (default: ``False``); if ``True``,
      do not raise weak errors (precision, abandon)

    TESTS::

        sage: from sage.rings.padics.padic_relaxed_errors import *
        sage: raise_error(0)
        sage: raise_error(64)
        Traceback (most recent call last):
        ...
        ZeroDivisionError: denominator is not invertible

        sage: raise_error(1)
        Traceback (most recent call last):
        ...
        PrecisionError: computation has been abandoned; try to increase precision

        sage: raise_error(1, permissive=True)
    """
    if error & ERROR_UNEXPECTED:
        raise RuntimeError("error code = %s" % error)
    if error & ERROR_CIRCULAR:
        raise RecursionError("definition looks circular")
    if error & ERROR_DIVISION:
        raise ZeroDivisionError("denominator is not invertible")
    if error & ERROR_INTEGRAL:
        raise ValueError("not in the ring of integers")
    if error & ERROR_NOTSQUARE:
        raise ValueError("not a square")
    if not permissive:
        if error & ERROR_OVERFLOW:
            raise OverflowError
        if error & (ERROR_PRECISION | ERROR_NOTDEFINED):
            raise PrecisionError("not enough precision")
        if error & ERROR_ABANDON:
            raise PrecisionError("computation has been abandoned; try to increase precision")



# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
from sage.libs.gsl.types cimport *

cdef extern from "gsl/gsl_sf_bessel.h":

  double  gsl_sf_bessel_J0(double x)

  int  gsl_sf_bessel_J0_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_J1(double x)

  int  gsl_sf_bessel_J1_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_Jn(int n, double x)

  int  gsl_sf_bessel_Jn_e(int n, double x, gsl_sf_result * result)

  int  gsl_sf_bessel_Jn_array(int nmin, int nmax, double x, double result_array[])

  double  gsl_sf_bessel_Y0(double x)

  int  gsl_sf_bessel_Y0_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_Y1(double x)

  int  gsl_sf_bessel_Y1_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_Yn(int n, double x)

  int  gsl_sf_bessel_Yn_e(int n, double x, gsl_sf_result * result)

  int  gsl_sf_bessel_Yn_array(int nmin, int nmax, double x, double result_array[])

  double  gsl_sf_bessel_I0(double x)

  int  gsl_sf_bessel_I0_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_I1(double x)

  int  gsl_sf_bessel_I1_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_In(int n, double x)

  int  gsl_sf_bessel_In_e(int n, double x, gsl_sf_result * result)

  int  gsl_sf_bessel_In_array(int nmin, int nmax, double x, double result_array[])

  double  gsl_sf_bessel_I0_scaled(double x)

  int  gsl_sf_bessel_I0_scaled_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_I1_scaled(double x)

  int  gsl_sf_bessel_I1_scaled_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_In_scaled(int n, double x)

  int  gsl_sf_bessel_In_scaled_e(int n, double x, gsl_sf_result * result)

  int  gsl_sf_bessel_In_scaled_array(int nmin, int nmax, double x, double result_array[])

  double  gsl_sf_bessel_K0(double x)

  int  gsl_sf_bessel_K0_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_K1(double x)

  int  gsl_sf_bessel_K1_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_Kn(int n, double x)

  int  gsl_sf_bessel_Kn_e(int n, double x, gsl_sf_result * result)

  int  gsl_sf_bessel_Kn_array(int nmin, int nmax, double x, double result_array[])

  double  gsl_sf_bessel_K0_scaled(double x)

  int  gsl_sf_bessel_K0_scaled_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_K1_scaled(double x)

  int  gsl_sf_bessel_K1_scaled_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_Kn_scaled(int n, double x)

  int  gsl_sf_bessel_Kn_scaled_e(int n, double x, gsl_sf_result * result)

  int  gsl_sf_bessel_Kn_scaled_array(int nmin, int nmax, double x, double result_array[])

  double  gsl_sf_bessel_j0(double x)

  int  gsl_sf_bessel_j0_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_j1(double x)

  int  gsl_sf_bessel_j1_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_j2(double x)

  int  gsl_sf_bessel_j2_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_jl(int l, double x)

  int  gsl_sf_bessel_jl_e(int l, double x, gsl_sf_result * result)

  int  gsl_sf_bessel_jl_array(int lmax, double x, double result_array[])

  int  gsl_sf_bessel_jl_steed_array(int lmax, double x, double * jl_x_array)

  double  gsl_sf_bessel_y0(double x)

  int  gsl_sf_bessel_y0_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_y1(double x)

  int  gsl_sf_bessel_y1_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_y2(double x)

  int  gsl_sf_bessel_y2_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_yl(int l, double x)

  int  gsl_sf_bessel_yl_e(int l, double x, gsl_sf_result * result)

  int  gsl_sf_bessel_yl_array(int lmax, double x, double result_array[])

  double  gsl_sf_bessel_i0_scaled(double x)

  int  gsl_sf_bessel_i0_scaled_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_i1_scaled(double x)

  int  gsl_sf_bessel_i1_scaled_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_i2_scaled(double x)

  int  gsl_sf_bessel_i2_scaled_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_il_scaled(int l, double x)

  int  gsl_sf_bessel_il_scaled_e(int l, double x, gsl_sf_result * result)

  int  gsl_sf_bessel_il_scaled_array(int lmax, double x, double result_array[])

  double  gsl_sf_bessel_k0_scaled(double x)

  int  gsl_sf_bessel_k0_scaled_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_k1_scaled(double x)

  int  gsl_sf_bessel_k1_scaled_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_k2_scaled(double x)

  int  gsl_sf_bessel_k2_scaled_e(double x, gsl_sf_result * result)

  double  gsl_sf_bessel_kl_scaled(int l, double x)

  int  gsl_sf_bessel_kl_scaled_e(int l, double x, gsl_sf_result * result)

  int  gsl_sf_bessel_kl_scaled_array(int lmax, double x, double result_array[])

  double  gsl_sf_bessel_Jnu(double nu, double x)

  int  gsl_sf_bessel_Jnu_e(double nu, double x, gsl_sf_result * result)

  int  gsl_sf_bessel_sequence_Jnu_e(double nu, gsl_mode_t mode, size_t size, double v[])

  double  gsl_sf_bessel_Ynu(double nu, double x)

  int  gsl_sf_bessel_Ynu_e(double nu, double x, gsl_sf_result * result)

  double  gsl_sf_bessel_Inu(double nu, double x)

  int  gsl_sf_bessel_Inu_e(double nu, double x, gsl_sf_result * result)

  double  gsl_sf_bessel_Inu_scaled(double nu, double x)

  int  gsl_sf_bessel_Inu_scaled_e(double nu, double x, gsl_sf_result * result)

  double  gsl_sf_bessel_Knu(double nu, double x)

  int  gsl_sf_bessel_Knu_e(double nu, double x, gsl_sf_result * result)

  double  gsl_sf_bessel_lnKnu(double nu, double x)

  int  gsl_sf_bessel_lnKnu_e(double nu, double x, gsl_sf_result * result)

  double  gsl_sf_bessel_Knu_scaled(double nu, double x)

  int  gsl_sf_bessel_Knu_scaled_e(double nu, double x, gsl_sf_result * result)

  double  gsl_sf_bessel_zero_J0(unsigned int s)

  int  gsl_sf_bessel_zero_J0_e(unsigned int s, gsl_sf_result * result)

  double  gsl_sf_bessel_zero_J1(unsigned int s)

  int  gsl_sf_bessel_zero_J1_e(unsigned int s, gsl_sf_result * result)

  double  gsl_sf_bessel_zero_Jnu(double nu, unsigned int s)

  int  gsl_sf_bessel_zero_Jnu_e(double nu, unsigned int s, gsl_sf_result * result)




# sage.doctest: needs sage.rings.finite_rings sage.schemes
"""
AG codes

Algebraic geometry codes or shortly AG codes are linear codes defined using
functions or differentials on algebraic curves over finite fields. Sage
implements evaluation AG codes and differential AG codes as Goppa defined in
[Gop1981]_ and provides decoding algorithms for them in full generality.

EXAMPLES::

    sage: k.<a> = GF(4)
    sage: A.<x,y> = AffineSpace(k, 2)
    sage: C = Curve(y^2 + y - x^3)
    sage: F = C.function_field()
    sage: pls = F.places()
    sage: p = C([0,0])
    sage: Q, = p.places()
    sage: pls.remove(Q)
    sage: G = 5*Q
    sage: codes.EvaluationAGCode(pls, G)
    [8, 5] evaluation AG code over GF(4)
    sage: codes.DifferentialAGCode(pls, G)
    [8, 3] differential AG code over GF(4)

As is well known, the two kinds of AG codes are dual to each other. ::

    sage: E = codes.EvaluationAGCode(pls, G)
    sage: D = codes.DifferentialAGCode(pls, G)
    sage: E.dual_code() == D
    True
    sage: D.dual_code() == E
    True

Decoders for both evaluation and differential AG codes are available.

.. toctree::

  ag_code_decoders

A natural generalization of classical Goppa codes is Cartier codes [Cou2014]_. Cartier codes are
subfield subcodes of differential AG codes.

EXAMPLES::

    sage: F.<a> = GF(9)
    sage: P.<x,y,z> = ProjectiveSpace(F, 2);
    sage: C = Curve(x^3*y + y^3*z + x*z^3)
    sage: F = C.function_field()
    sage: pls = F.places()
    sage: Z, = C([0,0,1]).places()
    sage: pls.remove(Z)
    sage: G = 3*Z
    sage: codes.CartierCode(pls, G)  # long time
    [9, 4] Cartier code over GF(3)

AUTHORS:

- Kwankyu Lee (2019-03): initial version
"""

# ****************************************************************************
#       Copyright (C) 2019 Kwankyu Lee <kwankyu@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.coding.ag_code_decoders import (
    DifferentialAGCodeEncoder,
    DifferentialAGCodeUniqueDecoder,
    EvaluationAGCodeEncoder,
    EvaluationAGCodeUniqueDecoder,
)
from sage.coding.linear_code import (
    AbstractLinearCode,
    LinearCodeGeneratorMatrixEncoder,
    LinearCodeSyndromeDecoder,
)
from sage.matrix.constructor import matrix
from sage.matrix.matrix_space import MatrixSpace
from sage.modules.free_module_element import vector
from sage.rings.function_field.place import FunctionFieldPlace


class AGCode(AbstractLinearCode):
    """
    Base class of algebraic geometry codes.

    A subclass of this class is required to define ``_function_field``
    attribute that refers to an abstract functiom field or the function field
    of the underlying curve used to construct a code of the class.
    """
    def base_function_field(self):
        """
        Return the function field used to construct the code.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: p = C([0,0])
            sage: Q, = p.places()
            sage: pls.remove(Q)
            sage: G = 5*Q
            sage: code = codes.EvaluationAGCode(pls, G)
            sage: code.base_function_field()
            Function field in y defined by y^2 + y + x^3
        """
        return self._function_field


class EvaluationAGCode(AGCode):
    """
    Evaluation AG code defined by rational places ``pls`` and a divisor ``G``.

    INPUT:

    - ``pls`` -- list of rational places of a function field

    - ``G`` -- a divisor whose support is disjoint from ``pls``

    If ``G`` is a place, then it is regarded as a prime divisor.

    EXAMPLES::

        sage: k.<a> = GF(4)
        sage: A.<x,y> = AffineSpace(k, 2)
        sage: C = Curve(y^2 + y - x^3)
        sage: F = C.function_field()
        sage: pls = F.places()
        sage: Q, = C.places_at_infinity()
        sage: pls.remove(Q)
        sage: G = 5*Q
        sage: codes.EvaluationAGCode(pls, G)
        [8, 5] evaluation AG code over GF(4)

        sage: G = F.get_place(5)
        sage: codes.EvaluationAGCode(pls, G)
        [8, 5] evaluation AG code over GF(4)
    """
    _registered_encoders = {}
    _registered_decoders = {}

    def __init__(self, pls, G):
        """
        Initialize.

        TESTS::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls = [pl for pl in pls if pl != Q]
            sage: G = 5*Q
            sage: code = codes.EvaluationAGCode(pls, G)
            sage: TestSuite(code).run()
        """
        if issubclass(type(G), FunctionFieldPlace):
            G = G.divisor()  # place is converted to a prime divisor

        F = G.parent().function_field()
        K = F.constant_base_field()
        n = len(pls)

        if any(p.degree() > 1 for p in pls):
            raise ValueError("there is a nonrational place among the places")

        if any(p in pls for p in G.support()):
            raise ValueError("the support of the divisor is not disjoint from the places")

        self._registered_encoders['evaluation'] = EvaluationAGCodeEncoder
        self._registered_decoders['K'] = EvaluationAGCodeUniqueDecoder

        super().__init__(K, n, default_encoder_name='evaluation',
                               default_decoder_name='K')

        # compute basis functions associated with a generator matrix
        basis_functions = G.basis_function_space()
        m = matrix([vector(K, [b.evaluate(p) for p in pls]) for b in basis_functions])
        I = MatrixSpace(K, m.nrows()).identity_matrix()
        mI = m.augment(I)
        mI.echelonize()
        M = mI.submatrix(0, 0, m.nrows(), m.ncols())
        T = mI.submatrix(0, m.ncols())
        r = M.rank()

        self._generator_matrix = M.submatrix(0, 0, r)
        self._basis_functions = [sum(c * b for c, b in zip(T[i], basis_functions))
                                     for i in range(r)]

        self._pls = tuple(pls)
        self._G = G

        self._function_field = F

    def __eq__(self, other):
        """
        Test equality of ``self`` with ``other``.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: codes.EvaluationAGCode(pls, 5*Q) == codes.EvaluationAGCode(pls, 6*Q)
            False
        """
        if self is other:
            return True

        if not isinstance(other, EvaluationAGCode):
            return False

        return self._pls == other._pls and self._G == other._G

    def __hash__(self):
        """
        Return the hash value of ``self``.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: code = codes.EvaluationAGCode(pls, 5*Q)
            sage: {code: 1}
            {[8, 5] evaluation AG code over GF(4): 1}
        """
        return hash((self._pls, self._G))

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: codes.EvaluationAGCode(pls, 7*Q)
            [8, 7] evaluation AG code over GF(4)
        """
        return "[{}, {}] evaluation AG code over GF({})".format(
                self.length(), self.dimension(), self.base_field().cardinality())

    def _latex_(self):
        r"""
        Return the latex representation of ``self``.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: code = codes.EvaluationAGCode(pls, 3*Q)
            sage: latex(code)
            [8, 3]\text{ evaluation AG code over }\Bold{F}_{2^{2}}
        """
        return r"[{}, {}]\text{{ evaluation AG code over }}{}".format(
                self.length(), self.dimension(), self.base_field()._latex_())

    def basis_functions(self):
        r"""
        Return the basis functions associated with the generator matrix.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: code = codes.EvaluationAGCode(pls, 3*Q)
            sage: code.basis_functions()
            (y + a*x + 1, y + x, (a + 1)*x)
            sage: matrix([[f.evaluate(p) for p in pls] for f in code.basis_functions()])
            [    1     0     0     1     a a + 1     1     0]
            [    0     1     0     1     1     0 a + 1     a]
            [    0     0     1     1     a     a a + 1 a + 1]
        """
        return tuple(self._basis_functions)

    def generator_matrix(self):
        r"""
        Return a generator matrix of the code.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: code = codes.EvaluationAGCode(pls, 3*Q)
            sage: code.generator_matrix()
            [    1     0     0     1     a a + 1     1     0]
            [    0     1     0     1     1     0 a + 1     a]
            [    0     0     1     1     a     a a + 1 a + 1]
        """
        return self._generator_matrix

    def designed_distance(self):
        """
        Return the designed distance of the AG code.

        If the code is of dimension zero, then a :exc:`ValueError` is raised.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: code = codes.EvaluationAGCode(pls, 3*Q)
            sage: code.designed_distance()
            5
        """
        if self.dimension() == 0:
            raise ValueError("not defined for zero code")

        d = self.length() - self._G.degree()
        return d if d > 0 else 1


class DifferentialAGCode(AGCode):
    """
    Differential AG code defined by rational places ``pls`` and a divisor ``G``.

    INPUT:

    - ``pls`` -- list of rational places of a function field

    - ``G`` -- a divisor whose support is disjoint from ``pls``

    If ``G`` is a place, then it is regarded as a prime divisor.

    EXAMPLES::

        sage: k.<a> = GF(4)
        sage: A.<x,y> = AffineSpace(k, 2)
        sage: C = A.curve(y^3 + y - x^4)
        sage: Q = C.places_at_infinity()[0]
        sage: O = C([0,0]).place()
        sage: pls = [p for p in C.places() if p not in [O, Q]]
        sage: G = -O + 3*Q
        sage: codes.DifferentialAGCode(pls, -O + Q)
        [3, 2] differential AG code over GF(4)

        sage: F = C.function_field()
        sage: G = F.get_place(1)
        sage: codes.DifferentialAGCode(pls, G)
        [3, 1] differential AG code over GF(4)
    """
    _registered_encoders = {}
    _registered_decoders = {}

    def __init__(self, pls, G):
        """
        Initialize.

        TESTS::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: code = codes.DifferentialAGCode(pls, 3*Q)
            sage: TestSuite(code).run()
        """
        if issubclass(type(G), FunctionFieldPlace):
            G = G.divisor()  # place is converted to a prime divisor

        F = G.parent().function_field()
        K = F.constant_base_field()
        n = len(pls)

        if any(p.degree() > 1 for p in pls):
            raise ValueError("there is a nonrational place among the places")

        if any(p in pls for p in G.support()):
            raise ValueError("the support of the divisor is not disjoint from the places")

        self._registered_encoders['residue'] = DifferentialAGCodeEncoder
        self._registered_decoders['K'] = DifferentialAGCodeUniqueDecoder

        super().__init__(K, n, default_encoder_name='residue',
                               default_decoder_name='K')

        # compute basis differentials associated with a generator matrix
        basis_differentials = (-sum(pls) + G).basis_differential_space()
        m = matrix([vector(K, [w.residue(p) for p in pls]) for w in basis_differentials])
        I = MatrixSpace(K, m.nrows()).identity_matrix()
        mI = m.augment(I)
        mI.echelonize()
        M = mI.submatrix(0, 0, m.nrows(), m.ncols())
        T = mI.submatrix(0, m.ncols())
        r = M.rank()

        self._generator_matrix = M.submatrix(0, 0, r)
        self._basis_differentials = [sum(c * w for c, w in zip(T[i], basis_differentials))
                                     for i in range(r)]

        self._pls = tuple(pls)
        self._G = G
        self._function_field = F

    def __eq__(self, other):
        """
        Test equality of ``self`` with ``other``.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: c1 = codes.DifferentialAGCode(pls, 3*Q)
            sage: c2 = codes.DifferentialAGCode(pls, 3*Q)
            sage: c1 is c2
            False
            sage: c1 == c2
            True
        """
        if self is other:
            return True

        if not isinstance(other, DifferentialAGCode):
            return False

        return self._pls == other._pls and self._G == other._G

    def __hash__(self):
        """
        Return the hash of ``self``.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: code = codes.DifferentialAGCode(pls, 3*Q)
            sage: {code: 1}
            {[8, 5] differential AG code over GF(4): 1}
        """
        return hash((self._pls, self._G))

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: codes.DifferentialAGCode(pls, 3*Q)
            [8, 5] differential AG code over GF(4)
        """
        return "[{}, {}] differential AG code over GF({})".format(
                self.length(), self.dimension(), self.base_field().cardinality())

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: code = codes.DifferentialAGCode(pls, 3*Q)
            sage: latex(code)
            [8, 5]\text{ differential AG code over }\Bold{F}_{2^{2}}
        """
        return r"[{}, {}]\text{{ differential AG code over }}{}".format(
                self.length(), self.dimension(), self.base_field()._latex_())

    def basis_differentials(self):
        r"""
        Return the basis differentials associated with the generator matrix.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: code = codes.DifferentialAGCode(pls, 3*Q)
            sage: matrix([[w.residue(p) for p in pls]
            ....:         for w in code.basis_differentials()])
            [    1     0     0     0     0 a + 1 a + 1     1]
            [    0     1     0     0     0 a + 1     a     0]
            [    0     0     1     0     0     a     1     a]
            [    0     0     0     1     0     a     0 a + 1]
            [    0     0     0     0     1     1     1     1]
        """
        return tuple(self._basis_differentials)

    def generator_matrix(self):
        """
        Return a generator matrix of the code.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: code = codes.DifferentialAGCode(pls, 3*Q)
            sage: code.generator_matrix()
            [    1     0     0     0     0 a + 1 a + 1     1]
            [    0     1     0     0     0 a + 1     a     0]
            [    0     0     1     0     0     a     1     a]
            [    0     0     0     1     0     a     0 a + 1]
            [    0     0     0     0     1     1     1     1]
        """
        return self._generator_matrix

    def designed_distance(self):
        """
        Return the designed distance of the differential AG code.

        If the code is of dimension zero, then a :exc:`ValueError` is raised.

        EXAMPLES::

            sage: k.<a> = GF(4)
            sage: A.<x,y> = AffineSpace(k, 2)
            sage: C = Curve(y^2 + y - x^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Q, = C.places_at_infinity()
            sage: pls.remove(Q)
            sage: code = codes.DifferentialAGCode(pls, 3*Q)
            sage: code.designed_distance()
            3
        """
        if self.dimension() == 0:
            raise ValueError("not defined for zero code")

        d = self._G.degree() - 2 * self._function_field.genus() + 2
        return d if d > 0 else 1


class CartierCode(AGCode):
    r"""
    Cartier code defined by rational places ``pls`` and a divisor ``G`` of a function field.

    INPUT:

    - ``pls`` -- list of rational places

    - ``G`` -- a divisor whose support is disjoint from ``pls``

    - ``r`` -- integer (default: 1)

    - ``name`` -- string; name of the generator of the subfield `\GF{p^r}`

    OUTPUT: cartier code over `\GF{p^r}` where `p` is the characteristic of the
    base constant field of the function field

    Note that if ``r`` is 1 the default, then ``name`` can be omitted.

    EXAMPLES::

        sage: F.<a> = GF(9)
        sage: P.<x,y,z> = ProjectiveSpace(F, 2);
        sage: C = Curve(x^3*y + y^3*z + x*z^3)
        sage: F = C.function_field()
        sage: pls = F.places()
        sage: Z, = C(0,0,1).places()
        sage: pls.remove(Z)
        sage: G = 3*Z
        sage: code = codes.CartierCode(pls, G)  # long time
        sage: code.minimum_distance()           # long time
        2
    """
    def __init__(self, pls, G, r=1, name=None):
        """
        Initialize.

        TESTS::

            sage: F.<a> = GF(9)
            sage: P.<x,y,z> = ProjectiveSpace(F, 2);
            sage: C = Curve(x^3*y + y^3*z + x*z^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Z, = C(0,0,1).places()
            sage: pls.remove(Z)
            sage: G = 3*Z
            sage: code = codes.CartierCode(pls, G)  # long time
            sage: TestSuite(code).run()             # long time
        """
        F = G.parent().function_field()
        K = F.constant_base_field()

        if any(p.degree() > 1 for p in pls):
            raise ValueError("there is a nonrational place among the places")

        if any(p in pls for p in G.support()):
            raise ValueError("the support of the divisor is not disjoint from the places")

        if K.degree() % r != 0:
            raise ValueError("{} does not divide the degree of the constant base field".format(r))

        n = len(pls)
        D = sum(pls)
        p = K.characteristic()

        subfield = K.subfield(r, name=name)

        # compute a basis R of the space of differentials in Omega(G - D)
        # fixed by the Cartier operator
        E = G - D

        Grp = E.parent()  # group of divisors
        V, fr_V, to_V = E.differential_space()

        EE = Grp(0)
        dic = E.dict()
        for place in dic:
            mul = dic[place]
            if mul > 0:
                mul = mul // p**r
            EE += mul * place

        W, fr_W, to_W = EE.differential_space()

        a = K.gen()
        field_basis = [a**i for i in range(K.degree())] # over prime subfield
        basis = E.basis_differential_space()

        m = []
        for w in basis:
            for c in field_basis:
                cw = F(c) * w # c does not coerce...
                carcw = cw
                for i in range(r): # apply cartier r times
                    carcw = carcw.cartier()
                m.append([f for e in to_W(carcw - cw) for f in vector(e)])

        ker = matrix(m).kernel()

        R = []
        s = len(field_basis)
        ncols = s * len(basis)
        for row in ker.basis():
            v = vector([K(row[d:d+s]) for d in range(0,ncols,s)])
            R.append(fr_V(v))

        # construct a generator matrix
        m = []
        col_index = D.support()
        for w in R:
            row = []
            for p in col_index:
                res = w.residue(p).trace() # lies in constant base field
                c = subfield(res) # as w is Cartier fixed
                row.append(c)
            m.append(row)

        self._generator_matrix = matrix(m).row_space().basis_matrix()

        self._pls = tuple(pls)
        self._G = G
        self._r = r
        self._function_field = F

        self._registered_encoders['GeneratorMatrix'] = LinearCodeGeneratorMatrixEncoder
        self._registered_decoders['Syndrome'] = LinearCodeSyndromeDecoder

        super().__init__(subfield, n,
                         default_encoder_name='GeneratorMatrix',
                         default_decoder_name='Syndrome')

    def __eq__(self, other):
        """
        Test equality of ``self`` with ``other``.

        EXAMPLES::

            sage: F.<a> = GF(9)
            sage: P.<x,y,z> = ProjectiveSpace(F, 2);
            sage: C = Curve(x^3*y + y^3*z + x*z^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Z, = C(0,0,1).places()
            sage: pls.remove(Z)
            sage: c1 = codes.CartierCode(pls, 3*Z)  # long time
            sage: c2 = codes.CartierCode(pls, 1*Z)  # long time
            sage: c1 == c2                          # long time
            False
        """
        if self is other:
            return True

        if not isinstance(other, CartierCode):
            return False

        return self._pls == other._pls and self._G == other._G and self._r == other._r

    def __hash__(self):
        """
        Return the hash of ``self``.

        EXAMPLES::

            sage: F.<a> = GF(9)
            sage: P.<x,y,z> = ProjectiveSpace(F, 2);
            sage: C = Curve(x^3*y + y^3*z + x*z^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Z, = C(0,0,1).places()
            sage: pls.remove(Z)
            sage: G = 3*Z
            sage: code = codes.CartierCode(pls, G)  # long time
            sage: {code: 1}                         # long time
            {[9, 4] Cartier code over GF(3): 1}
        """
        return hash((self._pls, self._G, self._r))

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: F.<a> = GF(9)
            sage: P.<x,y,z> = ProjectiveSpace(F, 2);
            sage: C = Curve(x^3*y + y^3*z + x*z^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Z, = C(0,0,1).places()
            sage: pls.remove(Z)
            sage: G = 3*Z
            sage: codes.CartierCode(pls, G)  # long time
            [9, 4] Cartier code over GF(3)
        """
        return "[{}, {}] Cartier code over GF({})".format(
                self.length(), self.dimension(), self.base_field().cardinality())

    def _latex_(self):
        r"""
        Return the latex representation of ``self``.

        EXAMPLES::

            sage: F.<a> = GF(9)
            sage: P.<x,y,z> = ProjectiveSpace(F, 2);
            sage: C = Curve(x^3*y + y^3*z + x*z^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Z, = C(0,0,1).places()
            sage: pls.remove(Z)
            sage: G = 3*Z
            sage: code = codes.CartierCode(pls, G)  # long time
            sage: latex(code)                       # long time
            [9, 4]\text{ Cartier code over }\Bold{F}_{3}
        """
        return r"[{}, {}]\text{{ Cartier code over }}{}".format(
                self.length(), self.dimension(), self.base_field()._latex_())

    def generator_matrix(self):
        r"""
        Return a generator matrix of the Cartier code.

        EXAMPLES::

            sage: F.<a> = GF(9)
            sage: P.<x,y,z> = ProjectiveSpace(F, 2);
            sage: C = Curve(x^3*y + y^3*z + x*z^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Z, = C(0,0,1).places()
            sage: pls.remove(Z)
            sage: G = 3*Z
            sage: code = codes.CartierCode(pls, G)  # long time
            sage: code.generator_matrix()           # long time
            [1 0 0 2 2 0 2 2 0]
            [0 1 0 2 2 0 2 2 0]
            [0 0 1 0 0 0 0 0 2]
            [0 0 0 0 0 1 0 0 2]
        """
        return self._generator_matrix

    def designed_distance(self):
        """
        Return the designed distance of the Cartier code.

        The designed distance is that of the differential code of which the
        Cartier code is a subcode.

        EXAMPLES::

            sage: F.<a> = GF(9)
            sage: P.<x,y,z> = ProjectiveSpace(F, 2);
            sage: C = Curve(x^3*y + y^3*z + x*z^3)
            sage: F = C.function_field()
            sage: pls = F.places()
            sage: Z, = C(0,0,1).places()
            sage: pls.remove(Z)
            sage: G = 3*Z
            sage: code = codes.CartierCode(pls, G)  # long time
            sage: code.designed_distance()          # long time
            1
        """
        if self.dimension() == 0:
            raise ValueError("not defined for zero code")

        d = self._G.degree() - 2 * self._function_field.genus() + 2
        return d if d > 0 else 1



r"""
Root system data for folded Cartan types

AUTHORS:

- Travis Scrimshaw (2013-01-12) - Initial version
"""
#*****************************************************************************
#       Copyright (C) 2013 Travis Scrimshaw <tscrim@ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.structure.sage_object import SageObject
from sage.structure.unique_representation import UniqueRepresentation
from sage.sets.family import Family
from sage.combinat.root_system.cartan_type import CartanType


class CartanTypeFolded(UniqueRepresentation, SageObject):
    r"""
    A Cartan type realized from a (Dynkin) diagram folding.

    Given a Cartan type `X`, we say `\hat{X}` is a folded Cartan
    type of `X` if there exists a diagram folding of the Dynkin
    diagram of `\hat{X}` onto `X`.

    A folding of a simply-laced Dynkin diagram `D` with index set `I` is an
    automorphism `\sigma` of `D` where all nodes any orbit of `\sigma` are not
    connected. The resulting Dynkin diagram `\hat{D}` is induced by
    `I / \sigma` where we identify edges in `\hat{D}` which are not incident
    and add a `k`-edge if we identify `k` incident edges and the arrow is
    pointing towards the indicent note. We denote the index set of `\hat{D}`
    by `\hat{I}`, and by abuse of notation, we denote the folding by `\sigma`.

    We also have scaling factors `\gamma_i` for `i \in \hat{I}` and defined
    as the unique numbers such that the map
    `\Lambda_j \mapsto \gamma_j \sum_{i \in \sigma^{-1}(j)} \Lambda_i`
    is the smallest proper embedding of the weight lattice of `X` to `\hat{X}`.

    If the Cartan type is simply laced, the default folding is the one
    induced from the identity map on `D`.

    If `X` is affine type, the default embeddings we consider here are:

    .. MATH::

        \begin{array}{ccl}
        C_n^{(1)}, A_{2n}^{(2)}, A_{2n}^{(2)\dagger}, D_{n+1}^{(2)}
        & \hookrightarrow & A_{2n-1}^{(1)}, \\
        A_{2n-1}^{(2)}, B_n^{(1)} & \hookrightarrow & D_{n+1}^{(1)}, \\
        E_6^{(2)}, F_4^{(1)} & \hookrightarrow & E_6^{(1)}, \\
        D_4^{(3)}, G_2^{(1)} & \hookrightarrow & D_4^{(1)},
        \end{array}

    and were chosen based on virtual crystals. In particular, the diagram
    foldings extend to crystal morphisms and gives a realization of
    Kirillov-Reshetikhin crystals for non-simply-laced types as simply-laced
    types. See [OSShimo03]_ and [FOS2009]_ for more details. Here we can compute
    `\gamma_i = \max(c) / c_i` where `(c_i)_i` are the translation factors
    of the root system. In a more type-dependent way, we can define `\gamma_i`
    as follows:

    1. There exists a unique arrow (multiple bond) in `X`.

       a. Suppose the arrow points towards 0. Then `\gamma_i = 1` for all
          `i \in I`.
       b. Otherwise `\gamma_i` is the order of `\sigma` for all `i` in the
          connected component of 0 after removing the arrow, else
          `\gamma_i = 1`.

    2. There is not a unique arrow. Thus `\hat{X} = A_{2n-1}^{(1)}` and
       `\gamma_i = 1` for all `1 \leq i \leq n-1`. If `i \in \{0, n\}`, then
       `\gamma_i = 2` if the arrow incident to `i` points away and is `1`
       otherwise.

    We note that `\gamma_i` only depends upon `X`.

    If the Cartan type is finite, then we consider the classical
    foldings/embeddings induced by the above affine foldings/embeddings:

    .. MATH::

        \begin{aligned}
        C_n & \hookrightarrow A_{2n-1}, \\
        B_n & \hookrightarrow D_{n+1}, \\
        F_4 & \hookrightarrow E_6, \\
        G_2 & \hookrightarrow D_4.
        \end{aligned}

    For more information on Cartan types, see
    :mod:`sage.combinat.root_system.cartan_type`.

    Other foldings may be constructed by passing in an optional
    ``folding_of`` second argument. See below.

    INPUT:

    - ``cartan_type`` -- the Cartan type `X` to create the folded type

    - ``folding_of`` -- the Cartan type `\hat{X}` which `X` is a folding of

    - ``orbit`` -- the orbit of the Dynkin diagram automorphism `\sigma`
      given as a list of lists where the `a`-th list corresponds to the `a`-th
      entry in `I` or a dictionary with keys in `I` and values as lists

    .. NOTE::

        If `X` is an affine type, we assume the special node is fixed
        under `\sigma`.

    EXAMPLES::

        sage: fct = CartanType(['C',4,1]).as_folding(); fct
        ['C', 4, 1] as a folding of ['A', 7, 1]
        sage: fct.scaling_factors()                                                     # needs sage.graphs
        Finite family {0: 2, 1: 1, 2: 1, 3: 1, 4: 2}
        sage: fct.folding_orbit()
        Finite family {0: (0,), 1: (1, 7), 2: (2, 6), 3: (3, 5), 4: (4,)}

    A simply laced Cartan type can be considered as a virtual type of
    itself::

        sage: fct = CartanType(['A',4,1]).as_folding(); fct
        ['A', 4, 1] as a folding of ['A', 4, 1]
        sage: fct.scaling_factors()                                                     # needs sage.graphs
        Finite family {0: 1, 1: 1, 2: 1, 3: 1, 4: 1}
        sage: fct.folding_orbit()
        Finite family {0: (0,), 1: (1,), 2: (2,), 3: (3,), 4: (4,)}

    Finite types::

        sage: fct = CartanType(['C',4]).as_folding(); fct
        ['C', 4] as a folding of ['A', 7]
        sage: fct.scaling_factors()
        Finite family {1: 1, 2: 1, 3: 1, 4: 2}
        sage: fct.folding_orbit()
        Finite family {1: (1, 7), 2: (2, 6), 3: (3, 5), 4: (4,)}

        sage: fct = CartanType(['F',4]).dual().as_folding(); fct
        ['F', 4] relabelled by {1: 4, 2: 3, 3: 2, 4: 1} as a folding of ['E', 6]
        sage: fct.scaling_factors()
        Finite family {1: 1, 2: 1, 3: 2, 4: 2}
        sage: fct.folding_orbit()
        Finite family {1: (1, 6), 2: (3, 5), 3: (4,), 4: (2,)}

    REFERENCES:

    - :wikipedia:`Dynkin_diagram#Folding`

    .. [OSShimo03] \M. Okado, A. Schilling, M. Shimozono.
       "Virtual crystals and fermionic formulas for type `D_{n+1}^{(2)}`,
       `A_{2n}^{(2)}`, and `C_n^{(1)}`". Representation Theory. **7** (2003).
       101-163. :doi:`10.1.1.192.2095`, :arxiv:`0810.5067`.
    """
    @staticmethod
    def __classcall_private__(cls, cartan_type, virtual, orbit):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: from sage.combinat.root_system.type_folded import CartanTypeFolded
            sage: sigma_list = [[0], [1,5], [2,4], [3]]
            sage: fct1 = CartanTypeFolded(['C',3,1], ['A',5,1], sigma_list)
            sage: sigma_tuple = tuple(map(tuple, sigma_list))
            sage: fct2 = CartanTypeFolded(CartanType(['C',3,1]), CartanType(['A',5,1]), sigma_tuple)
            sage: fct3 = CartanTypeFolded('C3~', 'A5~', {0:[0], 2:[2,4], 1:[1,5], 3:[3]})
            sage: fct1 is fct2 and fct2 is fct3
            True
        """
        if isinstance(cartan_type, CartanTypeFolded):
            return cartan_type
        cartan_type = CartanType(cartan_type)
        virtual = CartanType(virtual)
        if isinstance(orbit, dict):
            i_set = cartan_type.index_set()
            orb = [None] * len(i_set)
            for k, v in orbit.items():
                orb[i_set.index(k)] = tuple(v)
            orbit = tuple(orb)
        else:
            orbit = tuple(map(tuple, orbit))
        return super().__classcall__(cls, cartan_type, virtual, orbit)

    def __init__(self, cartan_type, folding_of, orbit):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: fct = CartanType(['C',4,1]).as_folding()
            sage: TestSuite(fct).run()
            sage: hash(fct)  # random
            42
        """
        self._cartan_type = cartan_type
        self._folding = folding_of
        self._orbit = orbit

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: CartanType(['C',4,1]).as_folding()
            ['C', 4, 1] as a folding of ['A', 7, 1]
        """
        return "{} as a folding of {}".format(self._cartan_type, self._folding)

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: fct = CartanType(['C', 4, 1]).as_folding()
            sage: latex(fct)
            C_{4}^{(1)} \hookrightarrow A_{7}^{(1)}
        """
        return self._cartan_type._latex_() + " \\hookrightarrow " + self._folding._latex_()

    def cartan_type(self):
        """
        Return the Cartan type of ``self``.

        EXAMPLES::

            sage: fct = CartanType(['C', 4, 1]).as_folding()
            sage: fct.cartan_type()
            ['C', 4, 1]
        """
        return self._cartan_type

    def folding_of(self):
        """
        Return the Cartan type of the virtual space.

        EXAMPLES::

            sage: fct = CartanType(['C', 4, 1]).as_folding()
            sage: fct.folding_of()
            ['A', 7, 1]
        """
        return self._folding

    @cached_method
    def folding_orbit(self):
        r"""
        Return the orbits under the automorphism `\sigma` as a
        dictionary (of tuples).

        EXAMPLES::

            sage: fct = CartanType(['C', 4, 1]).as_folding()
            sage: fct.folding_orbit()
            Finite family {0: (0,), 1: (1, 7), 2: (2, 6), 3: (3, 5), 4: (4,)}
        """
        return Family({i:tuple(self._orbit[pos])
                       for pos,i in enumerate(self._cartan_type.index_set())})

    @cached_method
    def scaling_factors(self):
        """
        Return the scaling factors of ``self``.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: fct = CartanType(['C', 4, 1]).as_folding()
            sage: fct.scaling_factors()
            Finite family {0: 2, 1: 1, 2: 1, 3: 1, 4: 2}
            sage: fct = CartanType(['BC', 4, 2]).as_folding()
            sage: fct.scaling_factors()
            Finite family {0: 1, 1: 1, 2: 1, 3: 1, 4: 2}
            sage: fct = CartanType(['BC', 4, 2]).dual().as_folding()
            sage: fct.scaling_factors()
            Finite family {0: 2, 1: 1, 2: 1, 3: 1, 4: 1}
            sage: CartanType(['BC', 4, 2]).relabel({0:4, 1:3, 2:2, 3:1, 4:0}).as_folding().scaling_factors()
            Finite family {0: 2, 1: 1, 2: 1, 3: 1, 4: 1}
        """
        if self._cartan_type.is_finite():
            L = self._cartan_type.root_system().ambient_space()

            def f(i):
                root = L.simple_root(i)
                coroot = L.simple_coroot(i)
                return root.leading_coefficient() / coroot.leading_coefficient()
            index_set = self._cartan_type.index_set()
            min_f = min(f(j) for j in index_set)
            return Family({i: int(f(i) / min_f) for i in index_set})
        if self._cartan_type.is_affine():
            c = self._cartan_type.translation_factors()
            cmax = max(c)
            return Family({i: int(cmax / c[i])
                           for i in self._cartan_type.index_set()})



"""
Extra functions for quadratic forms
"""

from sage.matrix.constructor import matrix
from sage.structure.element import Matrix
from sage.arith.misc import legendre_symbol
from sage.rings.integer_ring import ZZ


def is_triangular_number(n, return_value=False) -> bool | tuple:
    """
    Return whether ``n`` is a triangular number.

    A *triangular number* is a number of the form `k(k+1)/2` for some
    nonnegative integer `n`. See :wikipedia:`Triangular_number`. The sequence
    of triangular number is references as A000217 in the Online encyclopedia of
    integer sequences (OEIS).

    If you want to get the value of `k` for which `n=k(k+1)/2` set the
    argument ``return_value`` to ``True`` (see the examples below).

    INPUT:

    - ``n`` -- integer

    - ``return_value`` -- boolean (default: ``False``); if set to
      ``True`` the function returns a pair made of a boolean and the value `v`
      such that `v(v+1)/2 = n`

    EXAMPLES::

        sage: is_triangular_number(3)
        True
        sage: is_triangular_number(3, return_value=True)
        (True, 2)

        sage: is_triangular_number(2)
        False
        sage: is_triangular_number(2, return_value=True)
        (False, None)

        sage: is_triangular_number(25*(25+1)/2)
        True

        sage: is_triangular_number(10^6 * (10^6 +1)/2, return_value=True)
        (True, 1000000)

    TESTS::

        sage: F1 = [n for n in range(1,100*(100+1)/2)
        ....:       if is_triangular_number(n)]
        sage: F2 = [n*(n+1)/2 for n in range(1,100)]
        sage: F1 == F2
        True

        sage: for n in range(1000):
        ....:     res,v = is_triangular_number(n,return_value=True)
        ....:     assert res == is_triangular_number(n)
        ....:     if res: assert v*(v+1)/2 == n
    """
    n = ZZ(n)

    if return_value:
        if n < 0:
            return (False, None)
        if n == 0:
            return (True, ZZ.zero())
        s, r = (8 * n + 1).sqrtrem()
        if r:
            return (False, None)
        return (True, (s - 1) // 2)

    return (8 * n + 1).is_square()


def extend_to_primitive(A_input):
    r"""
    Given a matrix (resp. list of vectors), extend it to a square
    matrix (resp. list of vectors), such that its determinant is the
    gcd of its minors (i.e. extend the basis of a lattice to a
    "maximal" one in `\ZZ^n`).

    Author(s): Gonzalo Tornaria and Jonathan Hanke.

    INPUT:

    - ``A_input`` -- a matrix or a list of length n vectors (in the same space)

    OUTPUT: a square matrix or a list of n vectors (resp.)

    EXAMPLES::

        sage: A = Matrix(ZZ, 3, 2, range(6))
        sage: extend_to_primitive(A)
        [ 0  1 -1]
        [ 2  3  0]
        [ 4  5  0]

        sage: extend_to_primitive([vector([1,2,3])])
        [(1, 2, 3), (0, 1, 1), (-1, 0, 0)]
    """
    # Deal with a list of vectors
    if not isinstance(A_input, Matrix):
        A = matrix(A_input)      # Make a matrix A with the given rows.
        vec_output_flag = True
    else:
        A = A_input
        vec_output_flag = False

    # Arrange for A  to have more columns than rows.
    if A.is_square():
        return A
    if A.nrows() > A.ncols():
        return extend_to_primitive(A.transpose()).transpose()

    # Setup
    k = A.nrows()
    n = A.ncols()
    R = A.base_ring()

    # Smith normal form transformation, assuming more columns than rows
    V = A.smith_form()[2]

    # Extend the matrix in new coordinates, then switch back.
    B = A * V
    B_new = matrix(R, n - k, n)
    for i in range(n - k):
        B_new[i, n - i - 1] = 1
    C = B.stack(B_new)
    D = C * V**(-1)

    # Normalize for a positive determinant
    if D.det() < 0:
        D.rescale_row(n - 1, -1)

    # Return the current information
    if vec_output_flag:
        return D.rows()
    return D


def least_quadratic_nonresidue(p):
    r"""
    Return the smallest positive integer quadratic non-residue in `\ZZ/p\ZZ` for primes `p>2`.

    EXAMPLES::

        sage: least_quadratic_nonresidue(5)
        2
        sage: [least_quadratic_nonresidue(p) for p in prime_range(3, 100)]              # needs sage.libs.pari
        [2, 2, 3, 2, 2, 3, 2, 5, 2, 3, 2, 3, 2, 5, 2, 2, 2, 2, 7, 5, 3, 2, 3, 5]

    TESTS:

    This raises an error if input is a positive composite integer. ::

        sage: least_quadratic_nonresidue(20)
        Traceback (most recent call last):
        ...
        ValueError: p must be a prime number > 2

    This raises an error if input is 2. This is because every integer is a
    quadratic residue modulo 2. ::

        sage: least_quadratic_nonresidue(2)
        Traceback (most recent call last):
        ...
        ValueError: there are no quadratic non-residues in Z/2Z
    """
    p1 = abs(p)

    # Deal with the prime p = 2 and |p| <= 1.
    if p1 == 2:
        raise ValueError("there are no quadratic non-residues in Z/2Z")
    if p1 < 2:
        raise ValueError("p must be a prime number > 2")

    # Find the smallest non-residue mod p
    # For 7/8 of primes the answer is 2, 3 or 5:
    if p % 8 in (3, 5):
        return ZZ(2)
    if p % 12 in (5, 7):
        return ZZ(3)
    if p % 5 in (2, 3):
        return ZZ(5)
    # default case (first needed for p=71):
    if not p.is_prime():
        raise ValueError("p must be a prime number > 2")
    from sage.arith.srange import xsrange
    for r in xsrange(7, p):
        if legendre_symbol(r, p) == -1:
            return ZZ(r)



r"""
Interface to mwrank
"""

# ****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import os
import re
import weakref

from sage.interfaces.expect import Expect

instances = {}


def Mwrank(options='', server=None, server_tmpdir=None):
    """
    Create and return an mwrank interpreter, with given options.

    INPUT:

    - ``options`` -- string; passed when starting mwrank.
      The format is::

       -h       help            prints this info and quits
       -q       quiet           turns OFF banner display and prompt
       -v n     verbosity       sets verbosity to n (default=1)
       -o       PARI/GP output  turns ON extra PARI/GP short output (default: OFF)
       -p n     precision       sets precision to n decimals (default=15)
       -b n     quartic bound   bound on quartic point search (default=10)
       -x n     n aux           number of aux primes used for sieving (default=6)
       -l       list            turns ON listing of points (default: ON, unless v=0)
       -s       selmer_only     if set, computes Selmer rank only (default: not set)
       -d       skip_2nd_descent        if set, skips the second descent for curves with 2-torsion (default: not set)
       -S n     sat_bd          upper bound on saturation primes (default=100, -1 for automatic)

    .. WARNING::

       Do not use the option "-q" which turns off the prompt.

    EXAMPLES::

        sage: M = Mwrank('-v 0 -l')
        sage: print(M('0 0 1 -1 0'))
        Curve [0,0,1,-1,0] :    Rank = 1
        Generator 1 is [0:-1:1]; height 0.051...
        Regulator = 0.051...
    """
    try:
        X = instances[options]()
        if X:
            return X
    except KeyError:
        pass
    X = Mwrank_class(options, server=server, server_tmpdir=server_tmpdir)
    instances[options] = weakref.ref(X)
    return X


# regex matching '[a1,a2,a3,a4,a6]', no spaces, each ai a possibly signed integer
AINVS_LIST_RE = re.compile(r'\[[+-]?(\d+)(,[+-]?\d+){4}]')
# regex matching ' a1 a2 a3 a4 a6 ', any whitespace, each ai a possibly signed integer
AINVS_PLAIN_RE = re.compile(r'^(\s*)([+-]?(\d+)(\s+)){4}([+-]?(\d+))(\s*)$')


def validate_mwrank_input(s):
    r"""
    Return a string suitable for mwrank input, or raises an error.

    INPUT:

    - ``s`` -- one of the following:

        - a list or tuple of 5 integers [a1,a2,a3,a4,a6] or (a1,a2,a3,a4,a6)
        - a string of the form '[a1,a2,a3,a4,a6]' or 'a1 a2 a3 a4 a6' where a1, a2, a3, a4, a6 are integers

    OUTPUT:

    For valid input, a string of the form '[a1,a2,a3,a4,a6]'.
    For invalid input a :exc:`ValueError` is raised.

    EXAMPLES:

    A list or tuple of 5 integers::

        sage: from sage.interfaces.mwrank import validate_mwrank_input
        sage: validate_mwrank_input([1,2,3,4,5])
        '[1, 2, 3, 4, 5]'
        sage: validate_mwrank_input((-1,2,-3,4,-55))
        '[-1, 2, -3, 4, -55]'
        sage: validate_mwrank_input([1,2,3,4])
        Traceback (most recent call last):
        ...
        ValueError: [1, 2, 3, 4] is not valid input to mwrank (should have 5 entries)
        sage: validate_mwrank_input([1,2,3,4,i])
        Traceback (most recent call last):
        ...
        ValueError: [1, 2, 3, 4, I] is not valid input to mwrank (entries should be integers)


    A string of the form '[a1,a2,a3,a4,a6]' with any whitespace and integers ai::

        sage: validate_mwrank_input('0 -1 1 -7 6')
        '[0,-1,1,-7,6]'
        sage: validate_mwrank_input("[0,-1,1,0,0]\n")
        '[0,-1,1,0,0]'
        sage: validate_mwrank_input('0\t -1\t 1\t 0\t 0\n')
        '[0,-1,1,0,0]'
        sage: validate_mwrank_input('0 -1 1 -7 ')
        Traceback (most recent call last):
        ...
        ValueError: 0 -1 1 -7  is not valid input to mwrank
    """
    if isinstance(s, (list, tuple)):
        from sage.rings.integer_ring import ZZ
        if len(s) != 5:
            raise ValueError("%s is not valid input to mwrank (should have 5 entries)" % s)
        try:
            ai = [ZZ(a) for a in s]
            return str(ai)
        except (TypeError, ValueError):
            raise ValueError("%s is not valid input to mwrank (entries should be integers)" % s)

    if isinstance(s, str):
        if AINVS_PLAIN_RE.match(s):
            ai = s.split()
            return "[" + ",".join(ai) + "]"
        ss = s.replace(' ', '').replace('\n', '').replace('\t', '')
        if AINVS_LIST_RE.match(ss):
            return ss
    raise ValueError("%s is not valid input to mwrank" % s)


class Mwrank_class(Expect):
    """
    Interface to the Mwrank interpreter.
    """
    def __init__(self, options='', server=None, server_tmpdir=None):
        """
        INPUT:

        - ``options`` -- string; passed when starting mwrank.
          The format is::

           -h       help            prints this info and quits
           -q       quiet           turns OFF banner display and prompt
           -v n     verbosity       sets verbosity to n (default=1)
           -o       PARI/GP output  turns ON extra PARI/GP short output (default: OFF)
           -p n     precision       sets precision to n decimals (default=15)
           -b n     quartic bound   bound on quartic point search (default=10)
           -x n     n aux           number of aux primes used for sieving (default=6)
           -l       list            turns ON listing of points (default: ON, unless v=0)
           -s       selmer_only     if set, computes Selmer rank only (default: not set)
           -d       skip_2nd_descent        if set, skips the second descent for curves with 2-torsion (default: not set)
           -S n     sat_bd          upper bound on saturation primes (default=100, -1 for automatic)

        .. WARNING::

            Do not use the option "-q" which turns off the prompt.

        .. NOTE::

           Normally instances of this class would be created by
           calling the global function :meth:`Mwrank`.

        EXAMPLES::

            sage: from sage.interfaces.mwrank import Mwrank_class
            sage: M = Mwrank_class('-v 0 -l')
            sage: M('0 -1 1 0 0')
            'Curve [0,-1,1,0,0] :...Rank = 0...Regulator = 1...'

            sage: from sage.interfaces.mwrank import Mwrank_class
            sage: TestSuite(Mwrank_class).run()
        """
        Expect.__init__(self,
                        name='mwrank',
                        prompt='Enter curve: ',
                        command="mwrank %s" % options,
                        server=server,
                        server_tmpdir=server_tmpdir,
                        restart_on_ctrlc=True,
                        verbose_start=False)

    def __getattr__(self, attrname):
        """
        Standard function to return an attribute.

        EXAMPLES::

            sage: mwrank.zzz
            Traceback (most recent call last):
            ...
            AttributeError
        """
        raise AttributeError

    def __reduce__(self):
        """
        EXAMPLES::

            sage: Mwrank().__reduce__()
            (<function _reduce_load_mwrank at 0x...>, ())
        """

        return _reduce_load_mwrank, tuple([])

    def __call__(self, cmd):
        """
        Interface to eval method.

        INPUT:

        - ``cmd`` A string, or Sage object which when converted to a
          string gives valid input to ``mwrank``.  The conversion is
          done by :meth:`validate_mwrank_input`.

        EXAMPLES:

        The input can be five integers separated by whitespace::

            sage: mwrank('0 -1 1 0 0')
            'Curve [0,-1,1,0,0] :...Basic pair: I=16, J=-304...'

        Or a list or tuple of exactly five integers::

            sage: s = mwrank([0,-1,1,0,0])
            sage: "Rank = 0" in s and "been determined unconditionally" in s
            True

        TESTS:

        Invalid input raises an :exc:`ValueError` (see :issue:`10108`); this
        includes syntactically valid input which defines a singular curve::

            sage: mwrank(10)
            Traceback (most recent call last):
            ...
            ValueError: Invalid input: 10 is not valid input to mwrank

            sage: mwrank('0 0 0 0 0')
            Traceback (most recent call last):
            ...
            ValueError: Invalid input ([0,0,0,0,0]) to mwrank (singular curve)

            sage: mwrank('0 0 0 -3 2')
            Traceback (most recent call last):
            ...
            ValueError: Invalid input ([0,0,0,-3,2]) to mwrank (singular curve)
        """
        try:
            s = validate_mwrank_input(cmd)
        except ValueError as err:
            raise ValueError("Invalid input: %s" % err)
        try:
            return self.eval(s)
        except ValueError as err:
            raise ValueError(err)
        except RuntimeError:
            raise ValueError(cmd)

    def eval(self, s, **kwds):
        """
        Return mwrank's output for the given input.

        INPUT:

        - ``s`` -- string; a Sage object which when converted to a string
          gives valid input to ``mwrank``.  The conversion is done by
          :meth:`validate_mwrank_input`.  Possible formats are:

          - a string representing exactly five integers separated by
            whitespace, for example '1 2 3 4 5'

          - a string representing exactly five integers separated by
            commas, preceded by '[' and followed by ']' (with
            arbitrary whitespace), for example '[1 2 3 4 5]'

          - a list or tuple of exactly 5 integers.

        .. NOTE::

           If a :exc:`RuntimeError` exception is raised, then the mwrank
           interface is restarted and the command is retried once.

        EXAMPLES::

            sage: mwrank.eval('12 3 4 5 6')
            'Curve [12,3,4,5,6] :...'
            sage: mwrank.eval('[12, 3, 4, 5, 6]')
            'Curve [12,3,4,5,6] :...'
            sage: mwrank.eval([12, 3, 4, 5, 6])
            'Curve [12,3,4,5,6] :...'
            sage: mwrank.eval((12, 3, 4, 5, 6))
            'Curve [12,3,4,5,6] :...'
        """
        if self._expect is not None and not self._expect.isalive():
            # if mwrank is interrupted twice in rapid succession,
            # then it doesn't restart correctly, and we're left with:
            #   "RuntimeError: [Errno 9] Bad file descriptor"
            # Doing _start again fixes that always. See trac #5157.
            self._start()
        try:
            ss = validate_mwrank_input(s)
            return Expect.eval(self, ss, **kwds)
        except ValueError as err:
            raise ValueError('Invalid input: %s' % err)
        except RuntimeError:
            raise ValueError('Invalid input (%s) to mwrank (singular curve)' % s)

    def console(self):
        """
        Start the mwrank console.

        EXAMPLES::

            sage: mwrank.console() # not tested: expects console input
            Program mwrank: ...
        """
        mwrank_console()


# An instance
mwrank = Mwrank()


def _reduce_load_mwrank():
    """
    Return the standard mwrank instance.

    EXAMPLES::

        sage: from sage.interfaces.mwrank import _reduce_load_mwrank
        sage: _reduce_load_mwrank()
        Mwrank
    """
    return mwrank


def mwrank_console():
    """
    Start the mwrank console.

    EXAMPLES::

        sage: mwrank_console() # not tested: expects console input
        Program mwrank: ...
    """
    from sage.repl.rich_output.display_manager import get_display_manager
    if not get_display_manager().is_in_terminal():
        raise RuntimeError('Can use the console only in the terminal. Try %%mwrank magics instead.')
    os.system('mwrank')
